 /********************************************************************/
 /*                                                                  */
 /* Program name: MSGEXIT                                            */
 /*                                                                  */
 /* Description: Sample C channel message exit program that          */
 /*              processes reference messages.                       */
 /*                                                                  */
 /*  Statement:     Licensed Materials - Property of IBM             */                      
 /*                                                                  */      
 /*                 84H2000, 5765-B73                                */            
 /*                 84H2001, 6539-B42                                */            
 /*                 84H2002, 5765-B74                                */            
 /*                 84H2003, 5765-B75                                */            
 /*                 84H2004, 6539-B43                                */            
 /*                 94H5358, 5639-B69                                */
 /*                 (C) Copyright IBM Corp. 1994, 1997               */
 /*                                                                  */
 /********************************************************************/
 /*                                                                  */
 /* Function:                                                        */
 /*                                                                  */
 /*                                                                  */
 /*   MSGEXIT is a sample C program that is intended to be used as   */
 /*   a channel message exit.                                        */
 /*                                                                  */
 /*   It recognises reference messages with an object type that      */
 /*   matches the object type in the message exit user data field of */
 /*   the channel definition.                                        */
 /*                                                                  */
 /*   For these messages it does the following (other messages are   */
 /*   ignored) -                                                     */
 /*                                                                  */
 /*     Sender and Server channels  -                                */
 /*       The specified length of data is copied from the specified  */
 /*       offset of the specified file into the space remaining in   */
 /*       the agent buffer after the reference message.              */
 /*       If the end of the file is not reached the reference        */
 /*       message is put back on the transmission queue after        */
 /*       updating the DataLogicalOffset field.                      */
 /*                                                                  */
 /*     Requestor and Receiver channels  -                           */
 /*       If the DataLogicalOffset field is zero and the specified   */
 /*       file does not exist, it is created.                        */
 /*       The data following the reference message is added to the   */
 /*       end of the specified file.                                 */
 /*       If the reference message is not the last one for the       */
 /*       specified file, the reference message is discarded.        */
 /*       Otherwise, the reference message, without the appended     */
 /*       data, is returned to the channel program and will be put   */
 /*       to the target queue.                                       */
 /*                                                                  */
 /*   For Sender and Server channels, if the DataLogicalLength field */
 /*   in the input reference message is zero it is assumed that the  */
 /*   remaining part of the file, from DataLogicalOffset to the end  */
 /*   of the file, is to be sent along the channel.                  */
 /*   If it is non-zero, this indicates that only the length         */
 /*   specified is to be sent.                                       */
 /*                                                                  */
 /*   If an error occurs (e.g. unable to open a file),               */
 /*   MQCXP.ExitResponse is set to MQXCC_SUPPRESS_FUNCTION so that   */
 /*   the message being processed is not removed from the            */
 /*   transmission queue and the channels suspends execution.        */
 /*   A feedback code is returned in MQCXP.Feedback which will be    */
 /*   returned to the application that put the message, in the       */
 /*   MQMD.Feedback field, if the putting application requested      */
 /*   exception reports by setting MQRO_EXCEPTION in the             */
 /*   MQMD.Report field.                                             */
 /*                                                                  */
 /*   Program logic:                                                 */
 /*       Access the input parameters.                               */
 /*       Validate the fields in the reference message               */
 /*       Extract the object type from the message exit user data    */
 /*       field in the channel definition.                           */
 /*       If the message is a reference message and                  */
 /*          the object type matches the one from the MQCD           */
 /*         If the channel is a Sender or Server                     */
 /*           Check that the reference message is not followed by    */
 /*            data.                                                 */
 /*           Replace the input reference message with the converted */
 /*            one.                                                  */
 /*           Set the read length to the minimum of                  */
 /*            MQRMH.DataLogicalLength and the space remaining in    */
 /*            the agent buffer.                                     */
 /*           Copy data from specified offset of the specified file  */
 /*           into the space remaining at the end of the agent       */
 /*           buffer (up to the MaxMsgLength value for the channel). */
 /*           Update the DataLength parameter.                       */
 /*           If the end of the file has been reached or             */
 /*            the requested length has been read                    */
 /*             Set the MQRMHF_LAST flag in the reference message    */
 /*           Else                                                   */
 /*             Reset the MQRMHF_LAST flag in the message.           */
 /*             Update the DataLogicalOffset field in the message.   */
 /*             MQPUT1 the message to the transmission queue.        */
 /*         Else if the channel is a Receiver or Requestor           */
 /*           If DataLogicalOffset is zero and                       */
 /*              the specified file does not exist                   */
 /*             Create the file.                                     */
 /*           If DataLogicalOffset matches the current size of the   */
 /*           file                                                   */
 /*             Add the data to the end of the file.                 */
 /*           If the MQRMHF_LAST flag is set                         */
 /*             Replace the input reference message with the         */
 /*              converted one.                                      */
 /*             Decrement the DataLength parameter by the length of  */
 /*             the data.                                            */
 /*             Set DataLogicalOffset field in the MQRMH structure   */
 /*              to zero.                                            */
 /*             Set the MQRMH.DataLogicalLength to the size of the   */
 /*              file.                                               */
 /*           Else                                                   */
 /*             Set ExitResponse to MQXCC_SUPPRESS_FUNCTION          */
 /*             Set the MQMD.Report field to MQRO_DISCARD_MSG        */
 /*             Turn off the exception request flag in MQMD.Report   */
 /*                                                                  */
 /*   The fully qualified names of the objects, referred to in the   */
 /*   reference messages, are assumed to be not longer than 256      */
 /*   bytes.                                                         */
 /*                                                                  */
 /********************************************************************/
 /*                                                                  */
 /*   MsgExit has the usual parameters defined for channel exit      */
 /*       programs.                                                  */
 /*                                                                  */
 /********************************************************************/
 /*                                                                  */
 /*   If a function call within EXIT fails, a feedback code is       */
 /*   set to indicate the error. This is in addition to executing    */
 /*   a printf to output an error message.                           */
 /*   If the application that put the message to the remote          */
 /*   queue requested an exception report, the                       */
 /*   feedback code is returned in the exception report.             */
 /*   The feedback codes take the following form                     */
 /*     0x00ffeeee                                                   */
 /*   where ff is one of the XRMA_FB_xxx constants defined below and */
 /*   identifies the error e.g. fopen failed.                        */
 /*   eeee identifies the reason for the failure. This will be the   */
 /*   value of Reason in the case of MQ operations e.g. MQOPEN, or   */
 /*   the value of errno in the case of C functions like malloc and  */
 /*   fopen.                                                         */
 /*                                                                  */
 /********************************************************************/
 #include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <stddef.h>
 #include <string.h>
 #include <ctype.h>
 #include <stdarg.h>

    /* includes for MQI */
 #include <cmqc.h>
 #include <cmqxc.h>

 /********************************************************************/
 /* Trace output function                                            */
 /********************************************************************/

 char LOGBUF[257];

 void Aprintf(char *line)	
 {  	
 #ifdef _ALEX
	FILE *fp1;

	fp1= fopen("c:\\EXIT.LOG","a+");    
	fprintf(fp1,line);					
	fclose(fp1);
 #endif
 }

 void MQENTRY MsgExit(PMQCXP   pExitParms
                    ,PMQCD    pChannelDef
                    ,PMQLONG  pDataLength
                    ,PMQLONG  pAgentBufferLength
                    ,PMQCHAR  pAgentBuffer
                    ,PMQLONG  pExitBufferLength
                    ,PMQCHAR *pExitBuffer
                    );

 /********************************************************************/
 /* Typedefs.                                                        */
 /* The ExitUserArea within MQCXP is used to save the following      */
 /* values.                                                          */
 /********************************************************************/
 typedef struct tagSAVEDATA
 {
   MQHCONN HConn;       /* Connection handle       */
   MQLONG  ConnReason;  /* Reason code from MQCONN */
 } SAVEDATA, *PSAVEDATA;

 /********************************************************************/
 /* Constants                                                        */
 /********************************************************************/
 #define MAX_FILENAME_LENGTH 256
 #define NULL_POINTER        (void*)0L
 #define OK                  0L
 #define FAILED              1L
 #define MAX_MQRMH_LENGTH    1000

 /********************************************************************/
 /* Feedback codes                                                   */
 /********************************************************************/

 #define XRMA_FB_INVALIDEXITID       0x00110000
 #define XRMA_FB_INVALIDEXITREASON   0x00120000
 #define XRMA_FB_FOPENFILEFAILED     0x00130000
 #define XRMA_FB_FSEEKFILEFAILED     0x00140000
 #define XRMA_FB_FTELLFILEFAILED     0x00150000
 #define XRMA_FB_FREADFILEFAILED     0x00160000
 #define XRMA_FB_FWRITEFILEFAILED    0x00170000
 #define XRMA_FB_MQCONNFAILED        0x00180000
 #define XRMA_FB_MQOPENFAILED        0x00190000
 #define XRMA_FB_MQPUT1FAILED        0x001A0000
 #define XRMA_FB_MQGETFAILED         0x001B0000
 #define XRMA_FB_MQCLOSEFAILED       0x001C0000
 #define XRMA_FB_INVALIDFILEOFFSET   0x001D0000
 #define XRMA_FB_INVALIDCHANNELTYPE  0x001E0000
 #define XRMA_FB_INVALIDSTRUCLENGTH  0x001F0000
 #define XRMA_FB_MALLOCFAILED        0x00200000
 #define XRMA_FB_CONVERTFAILED       0x00210000
 #define XRMA_FB_INVALIDDATALENGTH   0x00220000

 /********************************************************************/
 /* Error and information messages                                   */
 /********************************************************************/
 #define STARTMSG             "AMQSXRM starting\n"
 #define ENDMSG               "AMQSXRM ending\n"
 #define REASONMSG            "Reason for calling exit is %s\n"
 #define INVALIDREASONMSG     "Reason %s is not valid for this exit\n"
 #define INVALIDEXITIDMSG     "Exitid %s is not valid for this exit\n"
 #define INVALIDMSGMSG      \
        "Message format is '%.8s'. Expecting '%.8s'. Message ignored\n"
 #define INVALIDOBJTYPEMSG  \
        "Object type is %s. Expecting %.8s. Message ignored\n"
 #define INVALIDOFFSETMSG   \
    "Data offset %d does not match file size %d for file %s."\
    " Message ignored\n"
 #define SRCENVERRORMSG   \
    "SrcEnv field not contained within MQRMH structure. "\
    "SrcEnvLength = %d. SrcEnvOffset = %d. StrucLength = %d."\
    " Message ignored\n"
 #define SRCNAMEERRORMSG   \
    "SrcName field not contained within MQRMH structure. "\
    "SrcNameLength = %d. SrcNameOffset = %d. StrucLength = %d."\
    " Message ignored\n"
 #define DESTENVERRORMSG   \
    "DestEnv field not contained within MQRMH structure. "\
    "DestEnvLength = %d. DestEnvOffset = %d. StrucLength = %d."\
    " Message ignored\n"
 #define DESTNAMEERRORMSG   \
    "DestName field not contained within MQRMH structure. "\
    "DestNameLength = %d. DestNameOffset = %d. StrucLength = %d."\
    " Message ignored\n"
 #define STRUCLENERRORMSG   \
    "Message does not contain a complete MQRMH structure. "\
    "StrucLength = %d. Message length = %d."\
    " Message ignored\n"
 #define DATALENERRORMSG   \
    "Message at Sender end contains data following the MQRMH structure. "\
    "Total length of headers = %d. Message length = %d."\
    " Message ignored\n"
 #define INVALIDCHLTYPEMSG    "Channel type is %d. Message ignored\n"
 #define UNKNOWNFILEMSG       "Cannot find file %s\n"
 #define CREATEFILEMSG        "File %s being created\n"
 #define UPDATEFILEMSG        "File %s being updated\n"
 #define READFILEMSG          "File %s being read\n"
 #define FILECOMPLETEMSG      "File %s is complete\n"
 #define EOFMSG               "End of file %s reached\n"
 #define OPENFAILEDMSG        "fopen to file %s failed. "
 #define SEEKFAILEDMSG        "fseek to file %s failed. "
 #define READFAILEDMSG        "fread to file %s failed. "
 #define TELLFAILEDMSG        "ftell to file %s failed. "
 #define WRITEFAILEDMSG       "fwrite to file %s failed. "
 #define MQCONNFAILEDMSG    \
            "MQCONN to queue manager %.48s failed with reason %d\n"
 #define MQOPENFAILEDMSG    \
            "MQOPEN to queue manager %.48s failed with reason %d\n"
 #define MQINQFAILEDMSG    \
            "MQINQ to queue manager %.48s failed with reason %d\n"
 #define MQPUT1FAILEDMSG    \
            "MQPUT1 to queue %.48s failed with reason %d\n"
 #define CONVERTMQRMHFAILEDMSG    \
            "ConvertMQRMH failed with return code %d\n"
 #define CONVERTDATAFAILEDMSG    \
            "ConvertBulkData failed with return code %d\n"
 #define ALLOCATEMQRMHFAILEDMSG    \
            "Unable to allocate a buffer to hold the converted MQRMH\n"
 #define ALLOCATEDATAHFAILEDMSG    \
  "Unable to allocate a buffer to hold the converted object data\n"


void MQENTRY MsgExit
                    (PMQCXP   pExitParms
                    ,PMQCD    pChannelDef
                    ,PMQLONG  pDataLength
                    ,PMQLONG  pAgentBufferLength
                    ,PMQCHAR  pAgentBuffer
                    ,PMQLONG  pExitBufferLength
                    ,PMQCHAR *pExitBuffer
                    )
{
  PMQXQH    pMQXQH;
  PMQRMH    pMQRMH;
  MQLONG    FileDataOffset;  /* offset of data within file           */
  MQLONG    FeedbackCode = 0;/* feedback code                        */
  char    * pMsgData;        /* pointer to data within reference msg */
  char      DestFileName[MAX_FILENAME_LENGTH+1];
                             /* Name of file at receiving end        */
  char      SourceFileName[MAX_FILENAME_LENGTH+1];
                             /* Name of file at sending end          */
  FILE    * fd = NULL_POINTER; /* file descriptor                    */
  char      mode[3];         /* file open mode                       */
  long      position;        /* position within file.                */
  int       rc = OK;         /* return code                          */
  size_t    itemswritten;    /* returned by fwrite                   */
  size_t    bytesread;       /* returned by fread                    */
  char      dummy;           /* work area                            */
  MQOD      od = {MQOD_DEFAULT};    /* Object Descriptor             */
  MQMD      md = {MQMD_DEFAULT};    /* Message Descriptor            */
  MQPMO     pmo = {MQPMO_DEFAULT};  /* put message options           */
  MQLONG    CompCode;               /* completion code               */
  MQLONG    Reason;                 /* reason code                   */
  MQLONG    MsgSpace;        /* Space remaining in input buffer      */
  MQLONG    ReadLength;      /* Number of bytes to read from file    */
  MQLONG    WriteLength;     /* Number of bytes to write to file     */
  int       WriteData=1;     /* Do we write the data                 */
  MQLONG    InputDataLength; /* Length requested in ref msg          */
  PSAVEDATA pSaveData = (PSAVEDATA)&(pExitParms -> ExitUserArea);
                             /* Save area for MQHCONN etc.           */
  MQLONG    OrigMQRMHLength; /* Length of MQRMH before conversion    */
  char      MsgBuffer[500];  /* message with inserts                 */

  char  ExitIds[][28] = {"MQXT_CHANNEL_SEC_EXIT"
                        ,"MQXT_CHANNEL_MSG_EXIT"
                        ,"MQXT_CHANNEL_SEND_EXIT"
                        ,"MQXT_CHANNEL_RCV_EXIT"
                        ,"MQXT_CHANNEL_MSG_RETRY_EXIT"
                        ,"MQXT_CHANNEL_AUTO_DEF_EXIT"
                        };
  char  ExitReasons[][14] = {"MQXR_INIT"
                            ,"MQXR_TERM"
                            ,"MQXR_MSG"
                            ,"MQXR_XMIT"
                            ,"MQXR_SEC_MSG"
                            ,"MQXR_INIT_SEC"
                            ,"MQXR_RETRY"
                            };

  pExitParms -> ExitResponse = MQXCC_OK;

  switch(pExitParms -> ExitId)
  {
    case MQXT_CHANNEL_MSG_EXIT:
         break;
    default:
         sprintf(LOGBUF,INVALIDEXITIDMSG,ExitIds[pExitParms->ExitId-
                                         MQXT_CHANNEL_SEC_EXIT]);
		 Aprintf(LOGBUF);
         FeedbackCode = XRMA_FB_INVALIDEXITID +
                        pExitParms->ExitId;
         goto MOD_EXIT;
         break;
  }

  switch(pExitParms -> ExitReason)
  {
    case MQXR_INIT:
         sprintf(LOGBUF,REASONMSG,ExitReasons[pExitParms->ExitReason-
                                         MQXR_INIT]);
		 Aprintf(LOGBUF);
         /************************************************************/
         /* Initialise the MQHCONN in ExitUserArea to show that we   */
         /* have not connected to the queue manager yet.             */
         /************************************************************/
         pSaveData -> HConn     = MQHO_UNUSABLE_HOBJ;
         goto MOD_EXIT;
         break;
    case MQXR_TERM:
         sprintf(LOGBUF,REASONMSG,ExitReasons[pExitParms->ExitReason-
                                         MQXR_INIT]);
		 Aprintf(LOGBUF);
         /************************************************************/
         /* If we have connected to a queue manager then disconnect. */
         /************************************************************/
         if (pSaveData -> HConn != MQHO_UNUSABLE_HOBJ &&
             pSaveData -> ConnReason != MQRC_ALREADY_CONNECTED)
         {
           MQDISC(&(pSaveData -> HConn)
                 ,&CompCode
                 ,&Reason
                 );
         }
         goto MOD_EXIT;
         break;
    case MQXR_MSG:
         sprintf(LOGBUF,REASONMSG,ExitReasons[pExitParms->ExitReason-
                                         MQXR_INIT]);
 		 Aprintf(LOGBUF);
        break;
    default:
         sprintf(LOGBUF,INVALIDREASONMSG,ExitReasons[pExitParms->ExitReason-
                                         MQXR_INIT]);
 		 Aprintf(LOGBUF);
         FeedbackCode = XRMA_FB_INVALIDEXITREASON +
                        pExitParms->ExitReason;
         goto MOD_EXIT;
         break;
  }

  /*******************************************************************/
  /* Get addresses of transmission header and message itself.        */
  /* The message should start with a reference message header.       */
  /*******************************************************************/

  pMQXQH  = (PMQXQH) pAgentBuffer;
  pMQRMH  = (PMQRMH) (pMQXQH + 1);

  /*******************************************************************/
  /* If the message format does not indicate a reference message     */
  /* then ignore the message.                                        */
  /*******************************************************************/
  if (memcmp(pMQXQH -> MsgDesc.Format
            ,MQFMT_REF_MSG_HEADER
            ,MQ_FORMAT_LENGTH
            ))
  {
    sprintf(LOGBUF,INVALIDMSGMSG,pMQXQH-> MsgDesc.Format,MQFMT_REF_MSG_HEADER);
	Aprintf(LOGBUF);
    
	goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Find the length of the original MQRMH.                          */
  /*******************************************************************/
  OrigMQRMHLength = pMQRMH -> StrucLength;

  /*******************************************************************/
  /* If the object type does not match the type specified in the     */
  /* channel definition then ignore the message.                     */
  /*******************************************************************/
  if (memcmp(pMQRMH -> ObjectType
            ,pExitParms -> ExitData
            ,sizeof(pMQRMH -> ObjectType)
            ))
  {
    sprintf(LOGBUF,INVALIDOBJTYPEMSG
          ,pMQRMH -> ObjectType
          ,pExitParms -> ExitData
          );
	Aprintf(LOGBUF);

    goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Check that the SrcEnv field lies within the structure           */
  /*******************************************************************/
  if (pMQRMH -> SrcEnvLength < 0 ||
      (pMQRMH -> SrcEnvLength > 0 &&
       (pMQRMH -> SrcEnvLength + pMQRMH -> SrcEnvOffset) >
       pMQRMH -> StrucLength
     ))
  {
     sprintf(LOGBUF,SRCENVERRORMSG,pMQRMH -> SrcEnvLength,
            pMQRMH -> SrcEnvOffset, pMQRMH -> StrucLength);
//     FeedbackCode = MQRC_SRC_ENV_ERROR;
     Aprintf(LOGBUF);
     
	 goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Check that the SrcName field lies within the structure          */
  /*******************************************************************/
  if (pMQRMH -> SrcNameLength < 0 ||
      (pMQRMH -> SrcNameLength > 0 &&
       (pMQRMH -> SrcNameLength + pMQRMH -> SrcNameOffset) >
       pMQRMH -> StrucLength
     ))
  {
     sprintf(LOGBUF,SRCNAMEERRORMSG,pMQRMH -> SrcNameLength,
            pMQRMH -> SrcNameOffset, pMQRMH -> StrucLength);
//     FeedbackCode = MQRC_SRC_NAME_ERROR;
	 Aprintf(LOGBUF);
     
	 goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Check that the DestEnv field lies within the structure          */
  /*******************************************************************/
  if (pMQRMH -> DestEnvLength < 0 ||
      (pMQRMH -> DestEnvLength > 0 &&
       (pMQRMH -> DestEnvLength + pMQRMH -> DestEnvOffset) >
       pMQRMH -> StrucLength
     ))
  {
     sprintf(LOGBUF,DESTENVERRORMSG,pMQRMH -> DestEnvLength,
            pMQRMH -> DestEnvOffset, pMQRMH -> StrucLength);
//     FeedbackCode = MQRC_DEST_ENV_ERROR;
	 Aprintf(LOGBUF);
     
	 goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Check that the DestName field lies within the structure         */
  /*******************************************************************/
  if (pMQRMH -> DestNameLength < 0 ||
      (pMQRMH -> DestNameLength > 0 &&
       (pMQRMH -> DestNameLength + pMQRMH -> DestNameOffset) >
       pMQRMH -> StrucLength
     ))
  {
     sprintf(LOGBUF,DESTNAMEERRORMSG,pMQRMH -> DestNameLength,
            pMQRMH -> DestNameOffset, pMQRMH -> StrucLength);
//    FeedbackCode = MQRC_DEST_NAME_ERROR;
 	 Aprintf(LOGBUF);
    
	 goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Check that the message contains a complete MQRMH structure      */
  /*******************************************************************/
  if (pMQRMH -> StrucLength > (MQLONG)(*pDataLength - sizeof(MQXQH)))
  {
     sprintf(LOGBUF,STRUCLENERRORMSG,pMQRMH -> StrucLength,*pDataLength);
	 Aprintf(LOGBUF);
     FeedbackCode = XRMA_FB_INVALIDSTRUCLENGTH;
     
	 goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Extract the following values from the reference message.        */
  /*  Offset of data within the file.                                */
  /*  Length of data to be read or written                           */
  /*  Name of file on destination system.                            */
  /*  Name of file on source system.                                 */
  /*******************************************************************/
  FileDataOffset  = pMQRMH -> DataLogicalOffset;
  InputDataLength = pMQRMH -> DataLogicalLength;
  memset(DestFileName,0,sizeof(DestFileName));
  memset(SourceFileName,0,sizeof(SourceFileName));
  memcpy(DestFileName
        ,(char*)pMQRMH + pMQRMH -> DestNameOffset
        ,pMQRMH -> DestNameLength
        );
  memcpy(SourceFileName
        ,(char*)pMQRMH + pMQRMH -> SrcNameOffset
        ,pMQRMH -> SrcNameLength
        );
  
  sprintf(LOGBUF,"\tFileOffset = %ld\n\tDataLength = %ld\n"
	     ,FileDataOffset, InputDataLength);
  Aprintf(LOGBUF);

  /*******************************************************************/
  /* Move the data to or from the file.                              */
  /*******************************************************************/
  switch (pChannelDef -> ChannelType)
  {
    /*****************************************************************/
    /* For server and sender channels append data from the file to   */
    /* the reference message and return it to the caller.            */
    /* If the end of the file is not reached then put the reference  */
    /* back on the transmission queue.                               */
    /*****************************************************************/
    case MQCHT_SENDER:
    case MQCHT_SERVER:

      /***************************************************************/
      /* Make sure there is no data following the MQRMH header.      */
      /* If there is, it may mean that a message with file data      */
      /* appended to it has been routed to a Sender channel instead  */
      /* of a Receiver channel.                                      */
      /***************************************************************/
      if (*pDataLength >
          (MQLONG)(sizeof(MQXQH) + OrigMQRMHLength)
         )
      {
         sprintf(LOGBUF,DATALENERRORMSG
               ,sizeof(MQXQH) + OrigMQRMHLength
               ,*pDataLength);
 		 Aprintf(LOGBUF);
         FeedbackCode = XRMA_FB_INVALIDDATALENGTH;
         
		 goto MOD_EXIT;
      }                           

      /***************************************************************/
      /* Calculate space remaining at end of input buffer and        */
      /* pointer to start of the space.                              */
      /* If the length of the agent buffer is greater than the       */
      /* MaxMsglength for the channel plus the transmission header   */
      /* size, then use the MaxMsgLength in the calculation.         */
      /* If InputDataLength > 0 this means that only that length     */
      /* is required. In this case, if the length specified is less  */
      /* than the space in the buffer then read just this length.    */
      /***************************************************************/
      MsgSpace  = ((*pAgentBufferLength >
                    (MQLONG)(pChannelDef->MaxMsgLength + sizeof(MQXQH))
                   )
                   ? (pChannelDef -> MaxMsgLength)
                   : (MQLONG)(*pAgentBufferLength - sizeof(MQXQH))
                  )
                  - pMQRMH -> StrucLength;
      pMsgData  = (PMQCHAR)pMQRMH + pMQRMH -> StrucLength;
      ReadLength = (InputDataLength > 0 &&
                    InputDataLength < MsgSpace)
                   ? InputDataLength
                   : MsgSpace;
      sprintf(LOGBUF,"\tMsgSpace = %ld\n\tReadLength = %ld\n",MsgSpace
		     , ReadLength);
      Aprintf(LOGBUF);
      sprintf(LOGBUF,"\tAgentBufferLength = %ld\n\tMQRMHLength = %ld\n"
		     ,*pAgentBufferLength, sizeof(MQRMH));
      Aprintf(LOGBUF);

      /***************************************************************/
      /* Open the file for reading                                   */
      /***************************************************************/
      fd = fopen(SourceFileName,"rb");

      if (fd == NULL_POINTER)
      {
        FeedbackCode = XRMA_FB_FOPENFILEFAILED + errno;
        sprintf(MsgBuffer,OPENFAILEDMSG,SourceFileName);
		Aprintf(MsgBuffer);
        perror(MsgBuffer);
        
		goto MOD_EXIT;
      }

      /***************************************************************/
      /* Position the file to the specified offset.                  */
      /***************************************************************/
      rc = fseek(fd
                ,FileDataOffset
                ,SEEK_SET
                );

      if (rc)
      {
        FeedbackCode = XRMA_FB_FSEEKFILEFAILED + errno;
        sprintf(MsgBuffer,SEEKFAILEDMSG,SourceFileName);
		Aprintf(MsgBuffer);
        perror(MsgBuffer);
        
		goto MOD_EXIT;
      }

      /***************************************************************/
      /* Read the data from the specified offset into the buffer     */
      /***************************************************************/
      bytesread = fread(pMsgData
                       ,1
                       ,ReadLength
                       ,fd
                       );

      if (ferror(fd))
      {
        FeedbackCode = XRMA_FB_FREADFILEFAILED + errno;
        sprintf(MsgBuffer,READFAILEDMSG,SourceFileName);
		Aprintf(MsgBuffer);
        perror(MsgBuffer);
        
		goto MOD_EXIT;
      }
      else
      if (feof(fd) ||
          (InputDataLength > 0  &&
           InputDataLength == (MQLONG)bytesread)
         )
      {
        /*************************************************************/
        /* End of file reached or all the requested data has been    */
        /* read.                                                     */
        /* Set MQRMHF_LAST flag.                                     */
        /*************************************************************/
        pMQRMH -> Flags  |= MQRMHF_LAST;
      }
      else
      {
        /*************************************************************/
        /* May have read last byte of the file (EOF is only set when */
        /* attempting to read beyond the last byte).                 */
        /* Try reading one more byte.                                */
        /*************************************************************/

        fread(&dummy,1,1,fd);

        if (ferror(fd))
        {
          FeedbackCode = XRMA_FB_FREADFILEFAILED + errno;
          sprintf(MsgBuffer,READFAILEDMSG,SourceFileName);
		  Aprintf(MsgBuffer);
          perror(MsgBuffer);
          
		  goto MOD_EXIT;
        }
        else
        if (feof(fd))
        {
          /***********************************************************/
          /* End of file reached before buffer was filled.           */
          /* Set MQRMHF_LAST flag.                                   */
          /***********************************************************/
          pMQRMH -> Flags |= MQRMHF_LAST;
        }
        else
        {
          /***********************************************************/
          /* End of file not reached.                                */
          /* Connect to the local queue manager if not already       */
          /* connected.                                              */
          /* Update the MQRMH.DataLogicalOffset field to identify    */
          /* the next piece of the file that is required.            */
          /* Decrement MQRMH.DataLength, if it is non-zero, by the   */
          /* number of bytes read.                                   */
          /* Turn off MQRMHF_LAST flag.                              */
          /* MQPUT1 the reference message back on the transmission   */
          /* queue.                                                  */
          /* If the message is persistent or the channel is not      */
          /* defined to be fast, then the message is put in          */
          /* syncpoint. Otherwise it is put out of syncpoint.        */
          /* The message is put with the same MQMD as was used in    */
          /* original MQPUT to the remote queue but the format is    */
          /* set to MQFMT_XMIT_Q_HEADER and Report set to none.      */
          /* Restore the current value of DataLogicalOffset before   */
          /* the message is returned to the caller.                  */
          /***********************************************************/

          if (pSaveData -> HConn == MQHO_UNUSABLE_HOBJ)
          {
            MQCONN(pChannelDef -> QMgrName
                  ,&(pSaveData -> HConn)
                  ,&CompCode
                  ,&(pSaveData -> ConnReason));

            if (CompCode == MQCC_FAILED)
            {
              FeedbackCode = XRMA_FB_MQCONNFAILED +
                             pSaveData -> ConnReason;
              sprintf(LOGBUF,MQCONNFAILEDMSG, pChannelDef -> QMgrName,
                     pSaveData -> ConnReason);
		      Aprintf(LOGBUF);
              
			  goto MOD_EXIT;
            }
          }

          strncpy(od.ObjectName
                 ,pChannelDef -> XmitQName
                 ,sizeof(od.ObjectName)
                 );
          strncpy(od.ObjectQMgrName
                 ,pChannelDef -> QMgrName
                 ,sizeof(od.ObjectQMgrName)
                 );

          memcpy((PMQCHAR)&md
                ,(PMQCHAR)&(pMQXQH -> MsgDesc)
                ,sizeof(pMQXQH -> MsgDesc)
                );
          memcpy(md.Format,MQFMT_XMIT_Q_HEADER,MQ_FORMAT_LENGTH);
          memcpy(md.MsgId   , MQMI_NONE, sizeof(md.MsgId));
          memcpy(md.CorrelId, MQCI_NONE, sizeof(md.CorrelId));

          md.Report =  MQRO_NONE;

          pmo.Options = ( //(pChannelDef -> NonPersistentMsgSpeed ==
                         // MQNPMS_NORMAL) ||
                         (pMQXQH -> MsgDesc.Persistence ==
                          MQPER_PERSISTENT))
                        ? MQPMO_FAIL_IF_QUIESCING |
                          MQPMO_SYNCPOINT
                        : MQPMO_FAIL_IF_QUIESCING |
                          MQPMO_NO_SYNCPOINT;

          pMQRMH -> Flags             &= ~MQRMHF_LAST;
          pMQRMH -> DataLogicalOffset += bytesread;

          if (pMQRMH -> DataLogicalLength > 0)
          {
            pMQRMH -> DataLogicalLength -= bytesread;
          }

          MQPUT1(pSaveData -> HConn  /* connection handle            */
                ,&od                 /* object descriptor for queue  */
                ,&md                 /* message descriptor           */
                ,&pmo                /* options                      */
                ,sizeof(MQXQH) +     /* message length               */
                 pMQRMH -> StrucLength
                ,pAgentBuffer        /* buffer                       */
                ,&CompCode           /* MQPUT1 completion code       */
                ,&Reason             /* reason code                  */
                );

          /***********************************************************/
          /* Restore original value of DataLogicalOffset.            */
          /***********************************************************/
          pMQRMH -> DataLogicalOffset -= bytesread;

          if (CompCode != MQCC_OK)
          {
            sprintf(LOGBUF,MQPUT1FAILEDMSG, od.ObjectName, Reason);
		    Aprintf(LOGBUF);
            FeedbackCode = XRMA_FB_MQPUT1FAILED + Reason;
            
			goto MOD_EXIT;
          }
        }
      }

      /***************************************************************/
      /* Set the MQRMH.DataLogicalLength field to the number of      */
      /* bytes read.                                                 */
      /* Set DataLength to the increased size of the message.        */
      /* The size of the message has increased by the number of      */
      /* bytes read from the file plus the difference in the sizes   */
      /* of the input and converted reference messages.              */
      /***************************************************************/
      pMQRMH -> DataLogicalLength  =  bytesread;
      *pDataLength                 += bytesread - OrigMQRMHLength +
                                      pMQRMH -> StrucLength;
      sprintf(LOGBUF,"\tDataLogicalLength = %ld\n",pMQRMH -> DataLogicalLength);
      Aprintf(LOGBUF);

      break;
    /*****************************************************************/
    /* For requestor and receiver channels, copy data from the       */
    /* reference message into the file.                              */
    /* If this is the first part of the file then (re)create the     */
    /* file first.                                                   */
    /* If this is the last part of the file then return the          */
    /* reference message to the caller, otherwise discard it.        */
    /*****************************************************************/
    case MQCHT_REQUESTER:
    case MQCHT_RECEIVER:

      /***************************************************************/
      /* Calculate address of data appended to MQRMH header.         */
      /* Calculate length of data to be written to the file.         */
      /***************************************************************/
      pMsgData  = (PMQCHAR)(pMQXQH +1) + OrigMQRMHLength;
      WriteLength = *pDataLength - sizeof(MQXQH) -
                    OrigMQRMHLength;

      /***************************************************************/
      /* If the data offset within the file is zero then the file is */
      /* created for writing. If it already exists, its contents are */
      /* destroyed.                                                  */
      /* Otherwise it is opened for appending.                       */
      /***************************************************************/
      strcpy(mode
            ,(FileDataOffset == 0) ? "wb" : "ab"
            );

      fd = fopen(DestFileName,mode);

      if (fd == NULL_POINTER)
      {
        FeedbackCode = XRMA_FB_FOPENFILEFAILED + errno;
        sprintf(MsgBuffer,OPENFAILEDMSG,DestFileName);
		Aprintf(MsgBuffer);
        perror(MsgBuffer);
        
		goto MOD_EXIT;
      }

      /***************************************************************/
      /* If the data offset within the file is not zero, check the   */
      /* current size of the file.                                   */
      /* If it matches the data offset, then write the data to the   */
      /* file.                                                       */
      /* Otherwise ignore the message.                               */
      /* If the data offset is zero then write the data to the file. */
      /***************************************************************/
      if (FileDataOffset != 0)
      {
                                       /* Go to the end of the file  */
        fseek(fd, 0, SEEK_END);
                                       /* Now, how big is it ?       */
        position = ftell(fd);

        if (position < 0)
        {
          FeedbackCode = XRMA_FB_FTELLFILEFAILED + errno;
          sprintf(MsgBuffer,TELLFAILEDMSG,DestFileName);
		  Aprintf(MsgBuffer);
          perror(MsgBuffer);
          
		  goto MOD_EXIT;
        }

        if (position > FileDataOffset)
        {
          /***********************************************************/
          /* File is bigger than the offset of this segment. This    */
          /* can happen during recovery when a channel terminates    */
          /* half way through a file transfer.                       */
          /* Our only response is to not write the data              */
          /***********************************************************/
          WriteData = 0;
        }

        if (position < FileDataOffset)
        {
          /***********************************************************/
          /* File is smaller than the offset of this segment. We     */
          /* must have lost a segment, alert the user to the error   */
          /***********************************************************/
          sprintf(LOGBUF,INVALIDOFFSETMSG,FileDataOffset,position,
                 DestFileName);
		  Aprintf(LOGBUF);
          FeedbackCode = XRMA_FB_INVALIDFILEOFFSET;
          
		  goto MOD_EXIT;
        }
      }

      /***************************************************************/
      /* Write the data to the specified offset in the file.         */
      /* The amount of data to be written is the length of the       */
      /* message minus the transmission and reference message headers*/
      /***************************************************************/
      if (WriteData && (WriteLength > 0))
      {
        itemswritten = fwrite(pMsgData
                             ,WriteLength
                             ,1
                             ,fd
                             );

        if (!itemswritten)
        {
          FeedbackCode = XRMA_FB_FWRITEFILEFAILED + errno;
          sprintf(MsgBuffer,WRITEFAILEDMSG,DestFileName);
          perror(MsgBuffer);
		  Aprintf(MsgBuffer);
          
		  goto MOD_EXIT;
        }
      }

      if (pMQRMH -> Flags == MQRMHF_LAST)
      {
        /*************************************************************/
        /* This is the last part of the file so we return the        */
        /* reference message to the caller without the appended      */
        /* object data, so set DataLength to the sum of the sizes of */
        /* the MQXQH and the output MQRMH.                           */
        /* The MQRMH is returned to the caller.                      */
        /* DataLogicalOffset is set to zero.                         */
        /* MQRMH.DataLogicalLength is set to the total size of the   */
        /* file.                                                     */
        /* ExitResponse is left at MQXCC_OK which means that the     */
        /* reference message (minus the appended data) will be put   */
        /* to the target queue.                                      */
        /*************************************************************/
        *pDataLength                 = sizeof(MQXQH) +
                                       pMQRMH -> StrucLength;
        pMQRMH -> DataLogicalLength  = FileDataOffset + WriteLength;
        pMQRMH -> DataLogicalOffset  = 0;
      }
      else
      {
        /*************************************************************/
        /* This is not the last part of the file so discard the      */
        /* message.                                                  */
        /*************************************************************/
        pExitParms -> ExitResponse =  MQXCC_SUPPRESS_FUNCTION;
        pMQXQH -> MsgDesc.Report   |= MQRO_DISCARD_MSG;
        pMQXQH -> MsgDesc.Report   &= ~MQRO_EXCEPTION;
      }

      break;
    /*****************************************************************/
    /* For other channel types, ignore the message.                  */
    /*****************************************************************/
    default:
      sprintf(LOGBUF,INVALIDCHLTYPEMSG
            ,pChannelDef -> ChannelType
            );
	  Aprintf(LOGBUF);
      
	  goto MOD_EXIT;
  }

  MOD_EXIT:
  
  /*******************************************************************/
  /* Close the file if necessary.                                    */
  /*******************************************************************/
  if (fd)
  {
    fclose(fd);
  }

  /*******************************************************************/
  /* If an error has occurred, stop the channel.                     */
  /* Set the feedback code to be returned in an exception report     */
  /* (if requested).                                                 */
  /*******************************************************************/
  if (FeedbackCode)
  {
    pExitParms -> ExitResponse =  MQXCC_SUPPRESS_FUNCTION;
    pExitParms -> Feedback     =  FeedbackCode;
  }

  return ;
}


#if MQAT_DEFAULT == MQAT_UNIX
 void MQStart(){;}
#endif
