 /********************************************************************/
 /*                                                                  */
 /* Program name: MSGEXIT                                            */
 /*                                                                  */
 /* Description: Sample C channel message exit program that          */
 /*              processes reference messages.                       */
 /*                                                                  */
 /*  Statement:     Licensed Materials - Property of IBM             */                      
 /*                                                                  */      
 /*                 84H2000, 5765-B73                                */            
 /*                 84H2001, 6539-B42                                */            
 /*                 84H2002, 5765-B74                                */            
 /*                 84H2003, 5765-B75                                */            
 /*                 84H2004, 6539-B43                                */            
 /*                 94H5358, 5639-B69                                */
 /*                 (C) Copyright IBM Corp. 1994, 1997               */
 /*                                                                  */
 /********************************************************************/
 /*                                                                  */
 /* Function:                                                        */
 /*                                                                  */
 /*                                                                  */
 /*   MSGEXIT is a sample C program that is intended to be used as   */
 /*   a channel message exit.                                        */
 /*                                                                  */
 /*   It recognises reference messages with an object type that      */
 /*   matches the object type in the message exit user data field of */
 /*   the channel definition.                                        */
 /*                                                                  */
 /*   For these messages it does the following (other messages are   */
 /*   ignored) -                                                     */
 /*                                                                  */
 /*     Sender and Server channels  -                                */
 /*       The specified length of data is copied from the specified  */
 /*       offset of the specified file into the space remaining in   */
 /*       the agent buffer after the reference message.              */
 /*       If the end of the file is not reached the reference        */
 /*       message is put back on the transmission queue after        */
 /*       updating the DataLogicalOffset field.                      */
 /*                                                                  */
 /*     Requestor and Receiver channels  -                           */
 /*       If the DataLogicalOffset field is zero and the specified   */
 /*       file does not exist, it is created.                        */
 /*       The data following the reference message is added to the   */
 /*       end of the specified file.                                 */
 /*       If the reference message is not the last one for the       */
 /*       specified file, the reference message is discarded.        */
 /*       Otherwise, the reference message, without the appended     */
 /*       data, is returned to the channel program and will be put   */
 /*       to the target queue.                                       */
 /*                                                                  */
 /*   For Sender and Server channels, if the DataLogicalLength field */
 /*   in the input reference message is zero it is assumed that the  */
 /*   remaining part of the file, from DataLogicalOffset to the end  */
 /*   of the file, is to be sent along the channel.                  */
 /*   If it is non-zero, this indicates that only the length         */
 /*   specified is to be sent.                                       */
 /*                                                                  */
 /*   If an error occurs (e.g. unable to open a file),               */
 /*   MQCXP.ExitResponse is set to MQXCC_SUPPRESS_FUNCTION so that   */
 /*   the message being processed is not removed from the            */
 /*   transmission queue and the channels suspends execution.        */
 /*   A feedback code is returned in MQCXP.Feedback which will be    */
 /*   returned to the application that put the message, in the       */
 /*   MQMD.Feedback field, if the putting application requested      */
 /*   exception reports by setting MQRO_EXCEPTION in the             */
 /*   MQMD.Report field.                                             */
 /*                                                                  */
 /*   Program logic:                                                 */
 /*       Access the input parameters.                               */
 /*       Validate the fields in the reference message               */
 /*       Extract the object type from the message exit user data    */
 /*       field in the channel definition.                           */
 /*       If the message is a reference message and                  */
 /*          the object type matches the one from the MQCD           */
 /*         If the channel is a Sender or Server                     */
 /*           Check that the reference message is not followed by    */
 /*            data.                                                 */
 /*           Replace the input reference message with the converted */
 /*            one.                                                  */
 /*           If this is not a continuation message                  */
 /*            Set the ObjectInstanceId = MsgId                      */
 /*           Set the read length to the minimum of                  */
 /*            MQRMH.DataLogicalLength and the space remaining in    */
 /*            the agent buffer.                                     */
 /*           Copy data from specified offset of the specified file  */
 /*           into the space remaining at the end of the agent       */
 /*           buffer (up to the MaxMsgLength value for the channel). */
 /*           Update the DataLength parameter.                       */
 /*           If the end of the file has been reached or             */
 /*            the requested length has been read                    */
 /*             Set the MQRMHF_LAST flag in the reference message    */
 /*           Else                                                   */
 /*             Set the MQRMHF_CONTINUATION flag in the message      */
 /*             Reset the MQRMHF_LAST flag in the message.           */
 /*             Update the DataLogicalOffset field in the message.   */
 /*             MQPUT1 the message to the transmission queue.        */
 /*         Else if the channel is a Receiver or Requestor           */
 /*           If DataLogicalOffset is zero and                       */
 /*              the specified file does not exist                   */
 /*             Create the file.                                     */
 /*           If DataLogicalOffset matches the current size of the   */
 /*           file                                                   */
 /*             Add the data to the end of the file.                 */
 /*           If the MQRMHF_LAST flag is set                         */
 /*             If there is a message on the Object State Queue for  */
 /*              this file                                           */
 /*               Remove message from queue                          */
 /*               Replace the input reference message with the       */
 /*                converted one.                                    */
 /*               Decrement the DataLength parameter by the length   */
 /*                of the data.                                      */
 /*               Set DataLogicalOffset field in the MQRMH structure */
 /*                to zero.                                          */
 /*               Set the MQRMH.DataLogicalLength to the size of the */
 /*                file.                                             */
 /*             Else                                                 */
 /*               Set ExitResponse to MQXCC_SUPPRESS_FUNCTION        */
 /*               Set the MQMD.Report field to MQRO_DISCARD_MSG      */
 /*               Turn off the exception request flag in MQMD.Report */
 /*           Else                                                   */
 /*             If the MQRMHF_CONTINUATION flag is not set           */
 /*               Place message on Object State Queue                */
 /*             Set ExitResponse to MQXCC_SUPPRESS_FUNCTION          */
 /*             Set the MQMD.Report field to MQRO_DISCARD_MSG        */
 /*             Turn off the exception request flag in MQMD.Report   */
 /*                                                                  */
 /*   The fully qualified names of the objects, referred to in the   */
 /*   reference messages, are assumed to be not longer than 256      */
 /*   bytes.                                                         */
 /*                                                                  */
 /********************************************************************/
 /*                                                                  */
 /*   MsgExit has the usual parameters defined for channel exit      */
 /*       programs.                                                  */
 /*                                                                  */
 /********************************************************************/
 /*                                                                  */
 /*   If a function call within EXIT fails, a feedback code is       */
 /*   set to indicate the error. This is in addition to executing    */
 /*   a printf to output an error message.                           */
 /*   If the application that put the message to the remote          */
 /*   queue requested an exception report, the                       */
 /*   feedback code is returned in the exception report.             */
 /*   The feedback codes take the following form                     */
 /*     0x00ffeeee                                                   */
 /*   where ff is one of the XRMA_FB_xxx constants defined below and */
 /*   identifies the error e.g. fopen failed.                        */
 /*   eeee identifies the reason for the failure. This will be the   */
 /*   value of Reason in the case of MQ operations e.g. MQOPEN, or   */
 /*   the value of errno in the case of C functions like malloc and  */
 /*   fopen.                                                         */
 /*                                                                  */
 /********************************************************************/
 /*                                                                  */
 /*                        IMPORTANT                                 */
 /*                        ---------                                 */
 /*                                                                  */
 /* There are a number of versions of the sample reference message   */
 /* exit.                                                            */
 /*                                                                  */
 /* AMQSXRMA.C   This version is shipped with MQSeries V5 for OS/2,  */
 /*              Windows NT, HP-UX, Sun Solaris and AIX and should   */
 /*              be used at the sending (and maybe the receiving     */
 /*              end) end of the channel when the sending channel is */
 /*              running under MQSeries V5.                          */
 /*              This is irrespective of which platform the          */
 /*              receiving channel is running on.                    */
 /*              The executable has the name AMQSXRM.DLL             */
 /*                                                                  */
 /* AMQSQRMA.C   This version is shipped with MQSeries V5 for OS/2,  */
 /*              Windows NT, HP-UX, Sun Solaris and AIX and can      */
 /*              be used at the sending and receiving ends of the    */
 /*              channel when the sending channel is NOT running     */
 /*              under MQSeries V5 on OS/2, Windows NT, HP-UX,       */
 /*              Sun Solaris or AIX.                                 */
 /*              For example, this version can be used when the      */
 /*              sending channel is running under                    */
 /*              MQSeries for AS/400 V4R2 or                         */
 /*              MQSeries for Windows V2.1                           */
 /*              This is irrespective of which platform the          */
 /*              receiving channel is running on.                    */
 /*              The executable has the name AMQSQRM.DLL             */
 /*                                                                  */
 /*              This version requires that a model queue, with the  */
 /*              name SYSTEM.DEFAULT.MODEL.PERMDYN.QUEUE and with a  */
 /*              queue definition type of PERMDYN, exists at the     */
 /*              receiving end.                                      */
 /*              The following MQSC command will create the queue.   */
 /*                                                                  */
 /*                 def qm(system.default.model.permdyn.queue) +     */
 /*                     deftype(permdyn)                             */
 /*                                                                  */
 /* MSGEXIT.C    This version can be used at the receiving end of    */
 /*              the channel when the sending channel is running     */
 /*              under MQSeries V5 on OS/2, Windows NT, HP-UX,       */
 /*              Sun Solaris or AIX.                                 */
 /*              For example, this version can be used when the      */
 /*              receiving channel is running under                  */
 /*              MQSeries for Windows V2.1 and the sending channel   */
 /*              is running under MQSeries V5.                       */
 /*              The executable has the name exit.dll                */
 /*                                                                  */
 /* MSGEXIT1.C   This version can be used at the sending and         */
 /*              receiving ends of the channel when the sending      */
 /*              channel is NOT running under MQSeries V5 on OS/2,   */
 /*              Windows NT, HP-UX, Sun Solaris or AIX.              */
 /*              For example, this version can be used when the      */
 /*              sending channel is running under                    */
 /*              MQSeries for Windows V2.1.                          */
 /*              This is irrespective of which platform the          */
 /*              receiving channel is running on.                    */
 /*              The executable has the name exit.dll                */
 /*                                                                  */
 /*              This version requires that a model queue, with the  */
 /*              name SYSTEM.DEFAULT.MODEL.PERMDYN.QUEUE and with a  */
 /*              queue definition type of PERMDYN, exists at the     */
 /*              receiving end.                                      */
 /*              The following MQSC command will create the queue.   */
 /*                                                                  */
 /*                 def qm(system.default.model.permdyn.queue) +     */
 /*                     deftype(permdyn)                             */
 /*                                                                  */
 /********************************************************************/
 #include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <stddef.h>
 #include <string.h>
 #include <ctype.h>
 #include <stdarg.h>

    /* includes for MQI */
 #include <cmqc.h>
 #include <cmqxc.h>

 /********************************************************************/
 /* Trace output function                                            */
 /********************************************************************/

 char LOGBUF[257];

 void Aprintf(char *line)	
 {  	
 #ifdef _ALEX
  FILE *fp1;

  fp1= fopen("c:\\EXIT.LOG","a+");    
  fprintf(fp1,line);					
  fclose(fp1);
 #endif
 }

 void MQENTRY MsgExit(PMQCXP   pExitParms
                    ,PMQCD    pChannelDef
                    ,PMQLONG  pDataLength
                    ,PMQLONG  pAgentBufferLength
                    ,PMQCHAR  pAgentBuffer
                    ,PMQLONG  pExitBufferLength
                    ,PMQCHAR *pExitBuffer
                    );

 /********************************************************************/
 /* Typedefs.                                                        */
 /* The ExitUserArea within MQCXP is used to save the following      */
 /* values.                                                          */
 /********************************************************************/
 typedef struct tagSAVEDATA
 {
   MQHCONN HConn;       /* Connection handle       */
   MQLONG  ConnReason;  /* Reason code from MQCONN */
 } SAVEDATA, *PSAVEDATA;

 /********************************************************************/
 /* Prototypes                                                       */
 /********************************************************************/
 int ConnectToQMgr(PSAVEDATA pSaveData
                  ,PMQCD     pChannelDef
                  ,PMQLONG   pFeedbackCode
                  ,char     *LOGBUF 
                  );

 int CheckObjectState(PSAVEDATA pSaveData
                     ,PMQCD     pChannelDef
                     ,PMQRMH    pMQRMH
                     ,PMQLONG   pFeedbackCode
                     ,char     *LOGBUF 
                     );

 int SetObjectState(PSAVEDATA pSaveData
                   ,PMQCD     pChannelDef
                   ,PMQRMH    pMQRMH
                   ,PMQLONG   pFeedbackCode
                   ,char     *LOGBUF 
                   );

 /********************************************************************/
 /* Constants                                                        */
 /********************************************************************/
 #define MAX_FILENAME_LENGTH 256
 #define NULL_POINTER        (void*)0L
 #define OK                  0L
 #define FAILED              1L
 #define OBJECT_FOUND        2L 
 #define OBJECT_NOT_FOUND    3L 
 #define MAX_MQRMH_LENGTH    1000
 #define MQRMHF_CONTINUATION 2     /* This is a continuation ref msg */
 #define STATE_QUEUE_NAME    "SYSTEM.REFMSG.STATE.QUEUE.%.20s"
 #define MODEL_PERMDYN_QUEUE "SYSTEM.DEFAULT.MODEL.PERMDYN.QUEUE"

 /********************************************************************/
 /* Feedback codes                                                   */
 /********************************************************************/

 #define XRMA_FB_INVALIDEXITID       0x00110000
 #define XRMA_FB_INVALIDEXITREASON   0x00120000
 #define XRMA_FB_FOPENFILEFAILED     0x00130000
 #define XRMA_FB_FSEEKFILEFAILED     0x00140000
 #define XRMA_FB_FTELLFILEFAILED     0x00150000
 #define XRMA_FB_FREADFILEFAILED     0x00160000
 #define XRMA_FB_FWRITEFILEFAILED    0x00170000
 #define XRMA_FB_MQCONNFAILED        0x00180000
 #define XRMA_FB_MQOPENFAILED        0x00190000
 #define XRMA_FB_MQPUT1FAILED        0x001A0000
 #define XRMA_FB_MQGETFAILED         0x001B0000
 #define XRMA_FB_MQCLOSEFAILED       0x001C0000
 #define XRMA_FB_INVALIDFILEOFFSET   0x001D0000
 #define XRMA_FB_INVALIDCHANNELTYPE  0x001E0000
 #define XRMA_FB_INVALIDSTRUCLENGTH  0x001F0000
 #define XRMA_FB_MALLOCFAILED        0x00200000
 #define XRMA_FB_CONVERTFAILED       0x00210000
 #define XRMA_FB_INVALIDDATALENGTH   0x00220000
 #define XRMA_FB_MODELMQOPENFAILED   0x00230000
 #define XRMA_FB_QWRONGTYPE          0x00240000
 #define XRMA_FB_PERMDYNMQOPENFAILED 0x00250000
 #define XRMA_FB_MQPUTSTATEFAILED    0x00260000

 /********************************************************************/
 /* Error and information messages                                   */
 /********************************************************************/
 #define STARTMSG             "AMQSXRM starting\n"
 #define ENDMSG               "AMQSXRM ending\n"
 #define REASONMSG            "Reason for calling exit is %s\n"
 #define INVALIDREASONMSG     "Reason %s is not valid for this exit\n"
 #define INVALIDEXITIDMSG     "Exitid %s is not valid for this exit\n"
 #define REPORTMSG            "Message is a report. Message ignored\n"
 #define INVALIDMSGMSG      \
        "Message format is '%.8s'. Expecting '%.8s'. Message ignored\n"
 #define INVALIDOBJTYPEMSG  \
        "Object type is %s. Expecting %.8s. Message ignored\n"
 #define INVALIDOFFSETMSG   \
    "Data offset %d does not match file size %d for file %s."\
    " Message ignored\n"
 #define SRCENVERRORMSG   \
    "SrcEnv field not contained within MQRMH structure. "\
    "SrcEnvLength = %d. SrcEnvOffset = %d. StrucLength = %d."\
    " Message ignored\n"
 #define SRCNAMEERRORMSG   \
    "SrcName field not contained within MQRMH structure. "\
    "SrcNameLength = %d. SrcNameOffset = %d. StrucLength = %d."\
    " Message ignored\n"
 #define DESTENVERRORMSG   \
    "DestEnv field not contained within MQRMH structure. "\
    "DestEnvLength = %d. DestEnvOffset = %d. StrucLength = %d."\
    " Message ignored\n"
 #define DESTNAMEERRORMSG   \
    "DestName field not contained within MQRMH structure. "\
    "DestNameLength = %d. DestNameOffset = %d. StrucLength = %d."\
    " Message ignored\n"
 #define STRUCLENERRORMSG   \
    "Message does not contain a complete MQRMH structure. "\
    "StrucLength = %d. Message length = %d."\
    " Message ignored\n"
 #define DATALENERRORMSG   \
    "Message at Sender end contains data following the MQRMH structure. "\
    "Total length of headers = %d. Message length = %d."\
    " Message ignored\n"
 #define INVALIDCHLTYPEMSG    "Channel type is %d. Message ignored\n"
 #define UNKNOWNFILEMSG       "Cannot find file %s\n"
 #define CREATEFILEMSG        "File %s being created\n"
 #define UPDATEFILEMSG        "File %s being updated\n"
 #define READFILEMSG          "File %s being read\n"
 #define FILECOMPLETEMSG      "File %s is complete\n"
 #define EOFMSG               "End of file %s reached\n"
 #define OPENFAILEDMSG        "fopen to file %s failed. "
 #define SEEKFAILEDMSG        "fseek to file %s failed. "
 #define READFAILEDMSG        "fread to file %s failed. "
 #define TELLFAILEDMSG        "ftell to file %s failed. "
 #define WRITEFAILEDMSG       "fwrite to file %s failed. "
 #define WRITEFAILEDMSG       "fwrite to file %s failed. "
 #define STATEMSGFOUNDMSG     "Message found on state queue\n "
 #define STATEMSGNOTFOUNDMSG  "Message NOT found on state queue\n "
 #define MSGPUTTOSTATEQMSG    "Message put to state queue\n "
 #define MQCONNFAILEDMSG    \
            "MQCONN to queue manager %.48s failed with reason %d\n"
 #define MQOPENFAILEDMSG    \
            "MQOPEN to queue manager %.48s failed with reason %d\n"
 #define MQINQFAILEDMSG    \
            "MQINQ to queue manager %.48s failed with reason %d\n"
 #define MQPUT1FAILEDMSG    \
            "MQPUT1 to queue %.48s failed with reason %d\n"
 #define MQPUTFAILEDMSG    \
            "MQPUT to queue %.48s failed with reason %d\n"
 #define MQOPENQFAILEDMSG    \
            "MQOPEN to queue %.48s failed with reason %d\n"
 #define QWRONGTYPEFAILEDMSG \
            "Model queue %s is not a permanent dynamic queue. " \
            "Its definition type is %d\n"
 #define MQCLOSEFAILEDMSG    \
            "MQCLOSE to queue %.48s failed with reason %d\n"
 #define CONVERTMQRMHFAILEDMSG    \
            "ConvertMQRMH failed with return code %d\n"
 #define CONVERTDATAFAILEDMSG    \
            "ConvertBulkData failed with return code %d\n"
 #define ALLOCATEMQRMHFAILEDMSG    \
            "Unable to allocate a buffer to hold the converted MQRMH\n"
 #define ALLOCATEDATAHFAILEDMSG    \
  "Unable to allocate a buffer to hold the converted object data\n"


void MQENTRY MsgExit
                    (PMQCXP   pExitParms
                    ,PMQCD    pChannelDef
                    ,PMQLONG  pDataLength
                    ,PMQLONG  pAgentBufferLength
                    ,PMQCHAR  pAgentBuffer
                    ,PMQLONG  pExitBufferLength
                    ,PMQCHAR *pExitBuffer
                    )
{
  PMQXQH    pMQXQH;
  PMQRMH    pMQRMH;
  MQLONG    FileDataOffset;  /* offset of data within file           */
  MQLONG    FeedbackCode = 0;/* feedback code                        */
  char    * pMsgData;        /* pointer to data within reference msg */
  char      DestFileName[MAX_FILENAME_LENGTH+1];
                             /* Name of file at receiving end        */
  char      SourceFileName[MAX_FILENAME_LENGTH+1];
                             /* Name of file at sending end          */
  FILE    * fd = NULL_POINTER; /* file descriptor                    */
  char      mode[3];         /* file open mode                       */
  long      position;        /* position within file.                */
  int       rc = OK;         /* return code                          */
  size_t    itemswritten;    /* returned by fwrite                   */
  size_t    bytesread;       /* returned by fread                    */
  char      dummy;           /* work area                            */
  MQOD      od = {MQOD_DEFAULT};    /* Object Descriptor             */
  MQMD      md = {MQMD_DEFAULT};    /* Message Descriptor            */
  MQPMO     pmo = {MQPMO_DEFAULT};  /* put message options           */
  MQLONG    CompCode;               /* completion code               */
  MQLONG    Reason;                 /* reason code                   */
  MQLONG    SavedFlags;      /* Save pMQRMH -> Flags                 */
  MQLONG    MsgSpace;        /* Space remaining in input buffer      */
  MQLONG    ReadLength;      /* Number of bytes to read from file    */
  MQLONG    WriteLength;     /* Number of bytes to write to file     */
  int       WriteData=1;     /* Do we write the data                 */
  MQLONG    InputDataLength; /* Length requested in ref msg          */
  PSAVEDATA pSaveData = (PSAVEDATA)&(pExitParms -> ExitUserArea);
                             /* Save area for MQHCONN etc.           */
  MQLONG    OrigMQRMHLength; /* Length of MQRMH before conversion    */
  char      MsgBuffer[500];  /* message with inserts                 */

  char  ExitIds[][28] = {"MQXT_CHANNEL_SEC_EXIT"
                        ,"MQXT_CHANNEL_MSG_EXIT"
                        ,"MQXT_CHANNEL_SEND_EXIT"
                        ,"MQXT_CHANNEL_RCV_EXIT"
                        ,"MQXT_CHANNEL_MSG_RETRY_EXIT"
                        ,"MQXT_CHANNEL_AUTO_DEF_EXIT"
                        };
  char  ExitReasons[][14] = {"MQXR_INIT"
                            ,"MQXR_TERM"
                            ,"MQXR_MSG"
                            ,"MQXR_XMIT"
                            ,"MQXR_SEC_MSG"
                            ,"MQXR_INIT_SEC"
                            ,"MQXR_RETRY"
                            };

  pExitParms -> ExitResponse = MQXCC_OK;

  switch(pExitParms -> ExitId)
  {
    case MQXT_CHANNEL_MSG_EXIT:
         break;
    default:
         sprintf(LOGBUF,INVALIDEXITIDMSG,ExitIds[pExitParms->ExitId-
                                         MQXT_CHANNEL_SEC_EXIT]);
         Aprintf(LOGBUF);
         FeedbackCode = XRMA_FB_INVALIDEXITID +
                        pExitParms->ExitId;
         goto MOD_EXIT;
         break;
  }

  switch(pExitParms -> ExitReason)
  {
    case MQXR_INIT:
         sprintf(LOGBUF,REASONMSG,ExitReasons[pExitParms->ExitReason-
                                         MQXR_INIT]);
         Aprintf(LOGBUF);
         /************************************************************/
         /* Initialise the MQHCONN in ExitUserArea to show that we   */
         /* have not connected to the queue manager yet.             */
         /************************************************************/
         pSaveData -> HConn     = MQHO_UNUSABLE_HOBJ;
         goto MOD_EXIT;
         break;
    case MQXR_TERM:
         sprintf(LOGBUF,REASONMSG,ExitReasons[pExitParms->ExitReason-
                                         MQXR_INIT]);
         Aprintf(LOGBUF);
         /************************************************************/
         /* If we have connected to a queue manager then disconnect. */
         /************************************************************/
         if (pSaveData -> HConn != MQHO_UNUSABLE_HOBJ &&
             pSaveData -> ConnReason != MQRC_ALREADY_CONNECTED)
         {
           MQDISC(&(pSaveData -> HConn)
                 ,&CompCode
                 ,&Reason
                 );
         }
         goto MOD_EXIT;
         break;
    case MQXR_MSG:
         sprintf(LOGBUF,REASONMSG,ExitReasons[pExitParms->ExitReason-
                                         MQXR_INIT]);
         Aprintf(LOGBUF);
        break;
    default:
         sprintf(LOGBUF,INVALIDREASONMSG,ExitReasons[pExitParms->ExitReason-
                                         MQXR_INIT]);
         Aprintf(LOGBUF);
         FeedbackCode = XRMA_FB_INVALIDEXITREASON +
                        pExitParms->ExitReason;
         goto MOD_EXIT;
         break;
  }

  /*******************************************************************/
  /* Get addresses of transmission header and message itself.        */
  /* The message should start with a reference message header.       */
  /*******************************************************************/

  pMQXQH  = (PMQXQH) pAgentBuffer;
  pMQRMH  = (PMQRMH) (pMQXQH + 1);

  /*******************************************************************/
  /* If the message format does not indicate a reference message     */
  /* then ignore the message.                                        */
  /*******************************************************************/
  if (memcmp(pMQXQH -> MsgDesc.Format
            ,MQFMT_REF_MSG_HEADER
            ,MQ_FORMAT_LENGTH
            ))
  {
    sprintf(LOGBUF,INVALIDMSGMSG,pMQXQH-> MsgDesc.Format,MQFMT_REF_MSG_HEADER);
    Aprintf(LOGBUF);
    
    goto MOD_EXIT;
  }

  /*******************************************************************/
  /* If the message is a report then ignore the message.             */
  /*******************************************************************/
  if (pMQXQH -> MsgDesc.MsgType == MQMT_REPORT)
  {
    sprintf(LOGBUF,REPORTMSG);
    Aprintf(LOGBUF);
    goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Find the length of the original MQRMH.                          */
  /*******************************************************************/
  OrigMQRMHLength = pMQRMH -> StrucLength;

  /*******************************************************************/
  /* If the object type does not match the type specified in the     */
  /* channel definition then ignore the message.                     */
  /*******************************************************************/
  if (memcmp(pMQRMH -> ObjectType
            ,pExitParms -> ExitData
            ,sizeof(pMQRMH -> ObjectType)
            ))
  {
    sprintf(LOGBUF,INVALIDOBJTYPEMSG
          ,pMQRMH -> ObjectType
          ,pExitParms -> ExitData
          );
    Aprintf(LOGBUF);

    goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Check that the SrcEnv field lies within the structure           */
  /*******************************************************************/
  if (pMQRMH -> SrcEnvLength < 0 ||
      (pMQRMH -> SrcEnvLength > 0 &&
       (pMQRMH -> SrcEnvLength + pMQRMH -> SrcEnvOffset) >
       pMQRMH -> StrucLength
     ))
  {
     sprintf(LOGBUF,SRCENVERRORMSG,pMQRMH -> SrcEnvLength,
            pMQRMH -> SrcEnvOffset, pMQRMH -> StrucLength);
//     FeedbackCode = MQRC_SRC_ENV_ERROR;
     Aprintf(LOGBUF);
     
     goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Check that the SrcName field lies within the structure          */
  /*******************************************************************/
  if (pMQRMH -> SrcNameLength < 0 ||
      (pMQRMH -> SrcNameLength > 0 &&
       (pMQRMH -> SrcNameLength + pMQRMH -> SrcNameOffset) >
       pMQRMH -> StrucLength
     ))
  {
     sprintf(LOGBUF,SRCNAMEERRORMSG,pMQRMH -> SrcNameLength,
            pMQRMH -> SrcNameOffset, pMQRMH -> StrucLength);
//     FeedbackCode = MQRC_SRC_NAME_ERROR;
     Aprintf(LOGBUF);
     
     goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Check that the DestEnv field lies within the structure          */
  /*******************************************************************/
  if (pMQRMH -> DestEnvLength < 0 ||
      (pMQRMH -> DestEnvLength > 0 &&
       (pMQRMH -> DestEnvLength + pMQRMH -> DestEnvOffset) >
       pMQRMH -> StrucLength
     ))
  {
     sprintf(LOGBUF,DESTENVERRORMSG,pMQRMH -> DestEnvLength,
            pMQRMH -> DestEnvOffset, pMQRMH -> StrucLength);
//     FeedbackCode = MQRC_DEST_ENV_ERROR;
     Aprintf(LOGBUF);
     
     goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Check that the DestName field lies within the structure         */
  /*******************************************************************/
  if (pMQRMH -> DestNameLength < 0 ||
      (pMQRMH -> DestNameLength > 0 &&
       (pMQRMH -> DestNameLength + pMQRMH -> DestNameOffset) >
       pMQRMH -> StrucLength
     ))
  {
     sprintf(LOGBUF,DESTNAMEERRORMSG,pMQRMH -> DestNameLength,
            pMQRMH -> DestNameOffset, pMQRMH -> StrucLength);
//    FeedbackCode = MQRC_DEST_NAME_ERROR;
     Aprintf(LOGBUF);
    
     goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Check that the message contains a complete MQRMH structure      */
  /*******************************************************************/
  if (pMQRMH -> StrucLength > (MQLONG)(*pDataLength - sizeof(MQXQH)))
  {
     sprintf(LOGBUF,STRUCLENERRORMSG,pMQRMH -> StrucLength,*pDataLength);
     Aprintf(LOGBUF);
     FeedbackCode = XRMA_FB_INVALIDSTRUCLENGTH;
     
     goto MOD_EXIT;
  }

  /*******************************************************************/
  /* Extract the following values from the reference message.        */
  /*  Offset of data within the file.                                */
  /*  Length of data to be read or written                           */
  /*  Name of file on destination system.                            */
  /*  Name of file on source system.                                 */
  /*******************************************************************/
  FileDataOffset  = pMQRMH -> DataLogicalOffset;
  InputDataLength = pMQRMH -> DataLogicalLength;
  memset(DestFileName,0,sizeof(DestFileName));
  memset(SourceFileName,0,sizeof(SourceFileName));
  memcpy(DestFileName
        ,(char*)pMQRMH + pMQRMH -> DestNameOffset
        ,pMQRMH -> DestNameLength
        );
  memcpy(SourceFileName
        ,(char*)pMQRMH + pMQRMH -> SrcNameOffset
        ,pMQRMH -> SrcNameLength
        );
  
  sprintf(LOGBUF,"\tFileOffset = %ld\n\tDataLength = %ld\n"
	     ,FileDataOffset, InputDataLength);
  Aprintf(LOGBUF);

  /*******************************************************************/
  /* Move the data to or from the file.                              */
  /*******************************************************************/
  switch (pChannelDef -> ChannelType)
  {
    /*****************************************************************/
    /* For server and sender channels append data from the file to   */
    /* the reference message and return it to the caller.            */
    /* If the end of the file is not reached then put the reference  */
    /* back on the transmission queue.                               */
    /*****************************************************************/
    case MQCHT_SENDER:
    case MQCHT_SERVER:

      /***************************************************************/
      /* Make sure there is no data following the MQRMH header.      */
      /* If there is, it may mean that a message with file data      */
      /* appended to it has been routed to a Sender channel instead  */
      /* of a Receiver channel.                                      */
      /***************************************************************/
      if (*pDataLength >
          (MQLONG)(sizeof(MQXQH) + OrigMQRMHLength)
         )
      {
         sprintf(LOGBUF,DATALENERRORMSG
               ,sizeof(MQXQH) + OrigMQRMHLength
               ,*pDataLength);
         Aprintf(LOGBUF);
         FeedbackCode = XRMA_FB_INVALIDDATALENGTH;
         
         goto MOD_EXIT;
      }                           

      /***************************************************************/
      /* If this is not a continuation message, copy the MsgId from  */
      /* the message descriptor to the ObjectInstanceId to give      */
      /* the transfer of this file a unique id.                      */
      /***************************************************************/
      if (!(pMQRMH -> Flags & MQRMHF_CONTINUATION))
      {
         memcpy(pMQRMH-> ObjectInstanceId
               ,pMQXQH -> MsgDesc.MsgId
               ,sizeof(pMQRMH -> ObjectInstanceId)
               );
      }

      /***************************************************************/
      /* Calculate space remaining at end of input buffer and        */
      /* pointer to start of the space.                              */
      /* If the length of the agent buffer is greater than the       */
      /* MaxMsglength for the channel plus the transmission header   */
      /* size, then use the MaxMsgLength in the calculation.         */
      /* If InputDataLength > 0 this means that only that length     */
      /* is required. In this case, if the length specified is less  */
      /* than the space in the buffer then read just this length.    */
      /***************************************************************/
      MsgSpace  = ((*pAgentBufferLength >
                    (MQLONG)(pChannelDef->MaxMsgLength + sizeof(MQXQH))
                   )
                   ? (pChannelDef -> MaxMsgLength)
                   : (MQLONG)(*pAgentBufferLength - sizeof(MQXQH))
                  )
                  - pMQRMH -> StrucLength;
      pMsgData  = (PMQCHAR)pMQRMH + pMQRMH -> StrucLength;
      ReadLength = (InputDataLength > 0 &&
                    InputDataLength < MsgSpace)
                   ? InputDataLength
                   : MsgSpace;
      sprintf(LOGBUF,"\tMsgSpace = %ld\n\tReadLength = %ld\n",MsgSpace
             , ReadLength);
      Aprintf(LOGBUF);
      sprintf(LOGBUF,"\tAgentBufferLength = %ld\n\tMQRMHLength = %ld\n"
             ,*pAgentBufferLength, sizeof(MQRMH));
      Aprintf(LOGBUF);

      /***************************************************************/
      /* Open the file for reading                                   */
      /***************************************************************/
      fd = fopen(SourceFileName,"rb");

      if (fd == NULL_POINTER)
      {
        FeedbackCode = XRMA_FB_FOPENFILEFAILED + errno;
        sprintf(MsgBuffer,OPENFAILEDMSG,SourceFileName);
        Aprintf(MsgBuffer);
        perror(MsgBuffer);
        
        goto MOD_EXIT;
      }

      /***************************************************************/
      /* Position the file to the specified offset.                  */
      /***************************************************************/
      rc = fseek(fd
                ,FileDataOffset
                ,SEEK_SET
                );

      if (rc)
      {
        FeedbackCode = XRMA_FB_FSEEKFILEFAILED + errno;
        sprintf(MsgBuffer,SEEKFAILEDMSG,SourceFileName);
        Aprintf(MsgBuffer);
        perror(MsgBuffer);
        
        goto MOD_EXIT;
      }

      /***************************************************************/
      /* Read the data from the specified offset into the buffer     */
      /***************************************************************/
      bytesread = fread(pMsgData
                       ,1
                       ,ReadLength
                       ,fd
                       );

      if (ferror(fd))
      {
        FeedbackCode = XRMA_FB_FREADFILEFAILED + errno;
        sprintf(MsgBuffer,READFAILEDMSG,SourceFileName);
        Aprintf(MsgBuffer);
        perror(MsgBuffer);
        
        goto MOD_EXIT;
      }
      else
      if (feof(fd) ||
          (InputDataLength > 0  &&
           InputDataLength == (MQLONG)bytesread)
         )
      {
        /*************************************************************/
        /* End of file reached or all the requested data has been    */
        /* read.                                                     */
        /* Set MQRMHF_LAST flag.                                     */
        /*************************************************************/
        pMQRMH -> Flags  |= MQRMHF_LAST;
      }
      else
      {
        /*************************************************************/
        /* May have read last byte of the file (EOF is only set when */
        /* attempting to read beyond the last byte).                 */
        /* Try reading one more byte.                                */
        /*************************************************************/

        fread(&dummy,1,1,fd);

        if (ferror(fd))
        {
          FeedbackCode = XRMA_FB_FREADFILEFAILED + errno;
          sprintf(MsgBuffer,READFAILEDMSG,SourceFileName);
          Aprintf(MsgBuffer);
          perror(MsgBuffer);
          
          goto MOD_EXIT;
        }
        else
        if (feof(fd))
        {
          /***********************************************************/
          /* End of file reached before buffer was filled.           */
          /* Set MQRMHF_LAST flag.                                   */
          /***********************************************************/
          pMQRMH -> Flags |= MQRMHF_LAST;
        }
        else
        {
          /***********************************************************/
          /* End of file not reached.                                */
          /* Connect to the local queue manager if not already       */
          /* connected.                                              */
          /* Update the MQRMH.DataLogicalOffset field to identify    */
          /* the next piece of the file that is required.            */
          /* Decrement MQRMH.DataLength, if it is non-zero, by the   */
          /* number of bytes read.                                   */
          /* Turn off MQRMHF_LAST and turn on MQRMHF_CONTINUATION    */
          /* flag.                                                   */
          /* MQPUT1 the reference message back on the transmission   */
          /* queue.                                                  */
          /* The message is put out of syncpoint.                    */
          /* The message is put with the same MQMD as was used in    */
          /* original MQPUT to the remote queue but the format is    */
          /* set to MQFMT_XMIT_Q_HEADER and Report set to none.      */
          /* Restore the current value of DataLogicalOffset and the  */
          /* continuation flag before the message is returned to the */
          /* caller.                                                 */
          /***********************************************************/

          rc = ConnectToQMgr(pSaveData
                            ,pChannelDef
                            ,&FeedbackCode
                            ,LOGBUF
                            );

          if (rc == FAILED)
          {
            goto MOD_EXIT;
          }

          strncpy(od.ObjectName
                 ,pChannelDef -> XmitQName
                 ,sizeof(od.ObjectName)
                 );
          strncpy(od.ObjectQMgrName
                 ,pChannelDef -> QMgrName
                 ,sizeof(od.ObjectQMgrName)
                 );

          memcpy((PMQCHAR)&md
                ,(PMQCHAR)&(pMQXQH -> MsgDesc)
                ,sizeof(pMQXQH -> MsgDesc)
                );
          memcpy(md.Format,MQFMT_XMIT_Q_HEADER,MQ_FORMAT_LENGTH);
          memcpy(md.MsgId   , MQMI_NONE, sizeof(md.MsgId));
          memcpy(md.CorrelId, MQCI_NONE, sizeof(md.CorrelId));

          md.Report =  MQRO_NONE;

          pmo.Options = MQPMO_FAIL_IF_QUIESCING |
                        MQPMO_NO_SYNCPOINT;

          pMQRMH -> Flags             &= ~MQRMHF_LAST;
          SavedFlags                   =  pMQRMH -> Flags;
          pMQRMH -> Flags             |=  MQRMHF_CONTINUATION;
          pMQRMH -> DataLogicalOffset += bytesread;

          if (pMQRMH -> DataLogicalLength > 0)
          {
            pMQRMH -> DataLogicalLength -= bytesread;
          }

          MQPUT1(pSaveData -> HConn  /* connection handle            */
                ,&od                 /* object descriptor for queue  */
                ,&md                 /* message descriptor           */
                ,&pmo                /* options                      */
                ,sizeof(MQXQH) +     /* message length               */
                 pMQRMH -> StrucLength
                ,pAgentBuffer        /* buffer                       */
                ,&CompCode           /* MQPUT1 completion code       */
                ,&Reason             /* reason code                  */
                );

          /***********************************************************/
          /* Restore original value of DataLogicalOffset and the     */
          /* continuation flag.                                      */ 
          /***********************************************************/
          pMQRMH -> DataLogicalOffset -= bytesread;
          pMQRMH -> Flags              = SavedFlags;

          if (CompCode != MQCC_OK)
          {
            sprintf(LOGBUF,MQPUT1FAILEDMSG, od.ObjectName, Reason);
            Aprintf(LOGBUF);
            FeedbackCode = XRMA_FB_MQPUT1FAILED + Reason;
            
            goto MOD_EXIT;
          }
        }
      }

      /***************************************************************/
      /* Set the MQRMH.DataLogicalLength field to the number of      */
      /* bytes read.                                                 */
      /* Set DataLength to the increased size of the message.        */
      /* The size of the message has increased by the number of      */
      /* bytes read from the file plus the difference in the sizes   */
      /* of the input and converted reference messages.              */
      /***************************************************************/
      pMQRMH -> DataLogicalLength  =  bytesread;
      *pDataLength                 += bytesread - OrigMQRMHLength +
                                      pMQRMH -> StrucLength;
      sprintf(LOGBUF,"\tDataLogicalLength = %ld\n"
             ,pMQRMH -> DataLogicalLength);
      Aprintf(LOGBUF);

      break;
    /*****************************************************************/
    /* For requestor and receiver channels, copy data from the       */
    /* reference message into the file.                              */
    /* If this is the first part of the file then (re)create the     */
    /* file first.                                                   */
    /* If this is the last part of the file then return the          */
    /* reference message to the caller, otherwise discard it.        */
    /*****************************************************************/
    case MQCHT_REQUESTER:
    case MQCHT_RECEIVER:

      /***************************************************************/
      /* Calculate address of data appended to MQRMH header.         */
      /* Calculate length of data to be written to the file.         */
      /***************************************************************/
      pMsgData  = (PMQCHAR)(pMQXQH +1) + OrigMQRMHLength;
      WriteLength = *pDataLength - sizeof(MQXQH) -
                    OrigMQRMHLength;

      /***************************************************************/
      /* If the data offset within the file is zero then the file is */
      /* created for writing. If it already exists, its contents are */
      /* destroyed.                                                  */
      /* Otherwise it is opened for appending.                       */
      /***************************************************************/
      strcpy(mode
            ,(FileDataOffset == 0) ? "wb" : "ab"
            );

      fd = fopen(DestFileName,mode);

      if (fd == NULL_POINTER)
      {
        FeedbackCode = XRMA_FB_FOPENFILEFAILED + errno;
        sprintf(MsgBuffer,OPENFAILEDMSG,DestFileName);
        Aprintf(MsgBuffer);
        perror(MsgBuffer);
        
        goto MOD_EXIT;
      }

      /***************************************************************/
      /* If the data offset within the file is not zero, check the   */
      /* current size of the file.                                   */
      /* If it matches the data offset, then write the data to the   */
      /* file.                                                       */
      /* Otherwise ignore the message.                               */
      /* If the data offset is zero then write the data to the file. */
      /***************************************************************/
      if (FileDataOffset != 0)
      {
                                       /* Go to the end of the file  */
        fseek(fd, 0, SEEK_END);
                                       /* Now, how big is it ?       */
        position = ftell(fd);

        if (position < 0)
        {
          FeedbackCode = XRMA_FB_FTELLFILEFAILED + errno;
          sprintf(MsgBuffer,TELLFAILEDMSG,DestFileName);
          Aprintf(MsgBuffer);
          perror(MsgBuffer);
          
          goto MOD_EXIT;
        }

        if (position > FileDataOffset)
        {
          /***********************************************************/
          /* File is bigger than the offset of this segment. This    */
          /* can happen during recovery when a channel terminates    */
          /* half way through a file transfer.                       */
          /* Our only response is to not write the data              */
          /***********************************************************/
          WriteData = 0;
        }

        if (position < FileDataOffset)
        {
          /***********************************************************/
          /* File is smaller than the offset of this segment. We     */
          /* must have lost a segment, alert the user to the error   */
          /***********************************************************/
          sprintf(LOGBUF,INVALIDOFFSETMSG,FileDataOffset,position,
                 DestFileName);
          Aprintf(LOGBUF);
          FeedbackCode = XRMA_FB_INVALIDFILEOFFSET;
          
          goto MOD_EXIT;
        }
      }

      /***************************************************************/
      /* Write the data to the specified offset in the file.         */
      /* The amount of data to be written is the length of the       */
      /* message minus the transmission and reference message headers*/
      /***************************************************************/
      if (WriteData && (WriteLength > 0))
      {
        itemswritten = fwrite(pMsgData
                             ,WriteLength
                             ,1
                             ,fd
                             );

        if (!itemswritten)
        {
          FeedbackCode = XRMA_FB_FWRITEFILEFAILED + errno;
          sprintf(MsgBuffer,WRITEFAILEDMSG,DestFileName);
          perror(MsgBuffer);
          Aprintf(MsgBuffer);
          
          goto MOD_EXIT;
        }
      }

      if (pMQRMH -> Flags & MQRMHF_LAST)
      {
        /*************************************************************/
        /* This is the last part of the file.                        */
        /* If this is the first time we have received this last part,*/
        /* we return the reference message to the caller without the */
        /* appended object data, so set DataLength to the sum of the */
        /* sizes of the MQXQH and the output MQRMH.                  */
        /* The MQRMH is returned to the caller.                      */
        /* DataLogicalOffset is set to zero.                         */
        /* MQRMH.DataLogicalLength is set to the total size of the   */
        /* file.                                                     */
        /* ExitResponse is left at MQXCC_OK which means that the     */
        /* reference message (minus the appended data) will be put   */
        /* to the target queue.                                      */
        /* If this last part has been returned to the caller before  */
        /* then discard this message.                                */ 
        /*************************************************************/
        rc = ConnectToQMgr(pSaveData
                          ,pChannelDef
                          ,&FeedbackCode
                          ,LOGBUF
                          );

        if (rc == FAILED)
        {
          goto MOD_EXIT;
        }

        rc = CheckObjectState(pSaveData
                             ,pChannelDef
                             ,pMQRMH
                             ,&FeedbackCode
                             ,LOGBUF
                             );

        switch(rc)
        {
          case OBJECT_FOUND:
            /*********************************************************/
            /* A message was found on the state queue for this object*/
            /* so this is the first time we have received this last  */
            /* piece so return the reference message to the caller   */
            /*********************************************************/ 
            *pDataLength                 = sizeof(MQXQH) +
                                           pMQRMH -> StrucLength;
            pMQRMH -> DataLogicalLength  = FileDataOffset+WriteLength;
            pMQRMH -> DataLogicalOffset  = 0;
            break;
          case OBJECT_NOT_FOUND:
            /*********************************************************/
            /* A message was not found on the state queue for this   */
            /* object so we assume that this last piece has been     */
            /* received before, so we can discard this message.      */
            /*********************************************************/ 
            pExitParms -> ExitResponse =  MQXCC_SUPPRESS_FUNCTION;
            pMQXQH -> MsgDesc.Report   |= MQRO_DISCARD_MSG;
            pMQXQH -> MsgDesc.Report   &= ~MQRO_EXCEPTION;
            break;
          default:
            goto MOD_EXIT;
        }
      }
      else
      {
        /*************************************************************/
        /* This is not the last part of the file so discard the      */
        /* message.                                                  */
        /* If this is the first part of the object, place a message, */
        /* with the CorrelId = ObjectInstanceId, onto the State      */
        /* Queue.                                                    */
        /*************************************************************/
        if (!(pMQRMH -> Flags & MQRMHF_CONTINUATION))
        {
          rc = ConnectToQMgr(pSaveData
                            ,pChannelDef
                            ,&FeedbackCode
                            ,LOGBUF
                            );

          if (rc == FAILED)
          {
            goto MOD_EXIT;
          }
            
          rc = SetObjectState(pSaveData
                             ,pChannelDef
                             ,pMQRMH
                             ,&FeedbackCode
                             ,LOGBUF
                             );
          if (rc == FAILED)
          {
            goto MOD_EXIT;
          }
        }  
        pExitParms -> ExitResponse =  MQXCC_SUPPRESS_FUNCTION;
        pMQXQH -> MsgDesc.Report   |= MQRO_DISCARD_MSG;
        pMQXQH -> MsgDesc.Report   &= ~MQRO_EXCEPTION;
      }

      break;
    /*****************************************************************/
    /* For other channel types, ignore the message.                  */
    /*****************************************************************/
    default:
      sprintf(LOGBUF,INVALIDCHLTYPEMSG
            ,pChannelDef -> ChannelType
            );
      Aprintf(LOGBUF);
      
      goto MOD_EXIT;
  }

  MOD_EXIT:
  
  /*******************************************************************/
  /* Close the file if necessary.                                    */
  /*******************************************************************/
  if (fd)
  {
    fclose(fd);
  }

  /*******************************************************************/
  /* If an error has occurred, stop the channel.                     */
  /* Set the feedback code to be returned in an exception report     */
  /* (if requested).                                                 */
  /*******************************************************************/
  if (FeedbackCode)
  {
    pExitParms -> ExitResponse =  MQXCC_SUPPRESS_FUNCTION;
    pExitParms -> Feedback     =  FeedbackCode;
  }

  return ;
}

/*********************************************************************/
/* ConnectToQMgr                                                     */
/*-------------------------------------------------------------------*/
/* If necessary, connect to the queue manager.                       */
/*********************************************************************/
int ConnectToQMgr(PSAVEDATA pSaveData
                 ,PMQCD     pChannelDef
                 ,PMQLONG   pFeedbackCode
                 ,char     *LOGBUF 
                 )
{
  int rc = OK;               /* Function return code                 */
  MQLONG   CompCode;

  /*******************************************************************/
  /* If not already connected, connect to the local queue manager    */
  /*******************************************************************/
  if (pSaveData -> HConn == MQHO_UNUSABLE_HOBJ)
  {
     MQCONN(pChannelDef -> QMgrName
           ,&(pSaveData -> HConn)
           ,&CompCode
           ,&(pSaveData -> ConnReason));

     if (CompCode == MQCC_FAILED)
     {
       sprintf(LOGBUF,MQCONNFAILEDMSG, pChannelDef -> QMgrName,
              pSaveData -> ConnReason);
       Aprintf(LOGBUF);
       *pFeedbackCode = XRMA_FB_MQCONNFAILED + pSaveData -> ConnReason;
       rc = FAILED;
     }
  }

  return rc;
}

/*********************************************************************/
/* CheckObjectState                                                  */
/*-------------------------------------------------------------------*/
/* Open the State Queue.                                             */
/* This is a permanent dynamic queue, created from the model queue   */
/* SYSTEM.DEFAULT_MODEL_PERMDYN queue, with the name                 */
/* SYSTEM.REFMSG.STATE.QUEUE.xxx where xxx is the 20 character       */
/* channel name.                                                     */
/* If the queue is successfully opened, an MQGET, specifying         */
/* a CorrelId = ObjectInstanceId from the reference message, is      */
/* issued .                                                          */
/*********************************************************************/
int CheckObjectState(PSAVEDATA pSaveData
                    ,PMQCD     pChannelDef
                    ,PMQRMH    pMQRMH
                    ,PMQLONG   pFeedbackCode
                    ,char     *LOGBUF 
                    )
{
  int rc = OK;               /* Function return code                 */
  MQLONG   CompCode;
  MQLONG   Reason;
  MQCHAR48 StateQueueName;
  MQOD     od = {MQOD_DEFAULT};     /* Object Descriptor             */
  MQMD     md = {MQMD_DEFAULT};     /* Message Descriptor            */
  MQGMO    gmo = {MQGMO_DEFAULT};   /* get message options           */
  MQHOBJ   Hobj = MQHO_UNUSABLE_HOBJ; /* object handle               */
  MQLONG   oo;                      /* MQOPEN options                */
  MQLONG   DataLength;              /* Length of message             */
  MQLONG   co  = MQCO_NONE;         /* MQCLOSE options               */

  /*******************************************************************/
  /* Determine name of state queue                                   */
  /*******************************************************************/
  sprintf(StateQueueName,STATE_QUEUE_NAME,pChannelDef->ChannelName);

  /*******************************************************************/
  /* Open the State Queue                                            */
  /*******************************************************************/
  strncpy(od.ObjectName
         ,StateQueueName
         ,(size_t)MQ_Q_NAME_LENGTH 
         );

  strncpy(od.ObjectQMgrName
         ,pChannelDef -> QMgrName
         ,sizeof(od.ObjectQMgrName)
         );
 
  oo = MQOO_FAIL_IF_QUIESCING +
       MQOO_INPUT_AS_Q_DEF;

  MQOPEN(pSaveData -> HConn          /* connection handle            */
        ,&od                         /* object descriptor for queue  */
        ,oo                          /* options                      */
        ,&Hobj                       /* object handle                */
        ,&CompCode                   /* MQOPEN completion code       */
        ,&Reason);                   /* reason code                  */

     
  if (Reason != MQRC_NONE)
  {
    rc = OBJECT_NOT_FOUND;
  }

  /*******************************************************************/
  /* Retrieve the state message for this object.                     */
  /* This is done by attempting to get a message,                    */
  /* with the CorrelId = ObjectInstanceId, from the state queue.     */
  /* If the get succeeds, then the transfer of this object is in     */
  /* progress.                                                       */
  /*******************************************************************/
  if (rc == OK)
  {
    gmo.Options = MQGMO_ACCEPT_TRUNCATED_MSG |
                  MQGMO_SYNCPOINT;

    memcpy(md.CorrelId
          ,pMQRMH -> ObjectInstanceId
          ,sizeof(md.CorrelId)
          ); 

    MQGET(pSaveData -> HConn
         ,Hobj
         ,&md
         ,&gmo
         ,0
         ,NULL_POINTER
         ,&DataLength
         ,&CompCode
         ,&Reason
         );      
              
    if (CompCode == MQCC_FAILED)
    {
      rc = OBJECT_NOT_FOUND;
      sprintf(LOGBUF,STATEMSGNOTFOUNDMSG);
      Aprintf(LOGBUF); 
    }
    else
    {
      rc = OBJECT_FOUND;
      sprintf(LOGBUF,STATEMSGFOUNDMSG);
      Aprintf(LOGBUF); 
    }
   
    /******************************************************************/
    /* Close the queue (if open)                                      */
    /******************************************************************/
    if (Hobj != MQHO_UNUSABLE_HOBJ)
    {
      MQCLOSE(pSaveData -> HConn
             ,&Hobj
             ,co
             ,&CompCode
             ,&Reason  
             );

      if (Reason != MQRC_NONE)
      {
        sprintf(LOGBUF,MQCLOSEFAILEDMSG,StateQueueName,Reason);
        Aprintf(LOGBUF);
      }  
    } 
  }

  return rc;
}

/*********************************************************************/
/* SetObjectState                                                    */
/*-------------------------------------------------------------------*/
/* Open the State Queue.                                             */
/* This is a permanent dynamic queue, created from the model queue   */
/* SYSTEM.DEFAULT_MODEL_PERMDYN queue, with the name                 */
/* SYSTEM.REFMSG.STATE.QUEUE.xxx where xxx is the 20 character       */
/* channel name.                                                     */
/* If the queue is successfully opened, an MQPUT, specifying         */
/* a CorrelId = ObjectInstanceId from the reference message, is      */
/* issued.                                                           */
/*********************************************************************/
int SetObjectState(PSAVEDATA pSaveData
                  ,PMQCD     pChannelDef
                  ,PMQRMH    pMQRMH
                  ,PMQLONG   pFeedbackCode
                  ,char     *LOGBUF 
                  )
{
  int rc = OK;               /* Function return code                 */
  MQLONG   CompCode;
  MQLONG   Reason;
  MQCHAR48 StateQueueName;
  MQOD     od = {MQOD_DEFAULT};     /* Object Descriptor             */
  MQOD     mod = {MQOD_DEFAULT};    /* Object Descriptor             */
  MQMD     md = {MQMD_DEFAULT};     /* Message Descriptor            */
  MQPMO    pmo = {MQPMO_DEFAULT};   /* put message options           */
  MQHOBJ   Hobj = MQHO_UNUSABLE_HOBJ; /* object handle               */
  MQLONG   oo;                      /* MQOPEN options                */
  MQLONG   co  = MQCO_NONE;         /* MQCLOSE options               */
  MQLONG   Selectors[] = {MQIA_DEFINITION_TYPE};
  MQLONG   QueueDefinitionType[1];

  /*******************************************************************/
  /* Determine name of state queue                                   */
  /*******************************************************************/
  sprintf(StateQueueName,STATE_QUEUE_NAME,pChannelDef->ChannelName);

  /*******************************************************************/
  /* Open the State Queue                                            */
  /*******************************************************************/
  strncpy(od.ObjectName
         ,StateQueueName
         ,(size_t)MQ_Q_NAME_LENGTH 
         );

  strncpy(od.ObjectQMgrName
         ,pChannelDef -> QMgrName
         ,sizeof(od.ObjectQMgrName)
         );
 
  oo = MQOO_FAIL_IF_QUIESCING + 
       MQOO_INQUIRE           +
       MQOO_OUTPUT;

  MQOPEN(pSaveData -> HConn          /* connection handle            */
        ,&od                         /* object descriptor for queue  */
        ,oo                          /* options                      */
        ,&Hobj                       /* object handle                */
        ,&CompCode                   /* MQOPEN completion code       */
        ,&Reason);                   /* reason code                  */

     
  switch (Reason)
  {
    case MQRC_UNKNOWN_OBJECT_NAME:
     /****************************************************************/
     /* Open the model queue to create the permanent dynamic queue   */
     /****************************************************************/
     strncpy(mod.ObjectName
            ,MODEL_PERMDYN_QUEUE
            ,(size_t)MQ_Q_NAME_LENGTH 
            );

     strncpy(mod.DynamicQName
            ,StateQueueName
            ,(size_t)MQ_Q_NAME_LENGTH 
            );

     strncpy(mod.ObjectQMgrName
            ,pChannelDef -> QMgrName
            ,sizeof(mod.ObjectQMgrName)
            );
 
     oo = MQOO_FAIL_IF_QUIESCING +
          MQOO_INQUIRE           + 
          MQOO_OUTPUT;

     MQOPEN(pSaveData -> HConn       /* connection handle            */
           ,&mod                     /* object descriptor for queue  */
           ,oo                       /* options                      */
           ,&Hobj                    /* object handle                */
           ,&CompCode                /* MQOPEN completion code       */
           ,&Reason);                /* reason code                  */

     if (Reason != MQRC_NONE)
     {
       sprintf(LOGBUF,MQOPENQFAILEDMSG, MODEL_PERMDYN_QUEUE, Reason);
       Aprintf(LOGBUF);
       *pFeedbackCode = XRMA_FB_MODELMQOPENFAILED + Reason;
       rc = FAILED;
     }
     break;

    case MQRC_NONE:
     break;

    default:
     sprintf(LOGBUF,MQOPENQFAILEDMSG, StateQueueName, Reason);
     Aprintf(LOGBUF);
     *pFeedbackCode = XRMA_FB_PERMDYNMQOPENFAILED + Reason;
     rc = FAILED;
     break;
  }

  /*******************************************************************/
  /* Make sure the queue is permanent dynamic and not temporary      */
  /* dynamic.                                                        */
  /*******************************************************************/
  if (rc == OK)
  {
     MQINQ(pSaveData -> HConn
          ,Hobj
          ,1                         /* 1 selector                   */
          ,Selectors
          ,1                         /* 1 integer selector           */
          ,QueueDefinitionType
          ,0L                        /* Character attribute length   */
          ,NULL_POINTER
          ,&CompCode
          ,&Reason
          );
                                             
     if (Reason == MQRC_NONE)
     {
       if (QueueDefinitionType[0] != MQQDT_PERMANENT_DYNAMIC)
       {
         sprintf(LOGBUF,QWRONGTYPEFAILEDMSG, MODEL_PERMDYN_QUEUE 
                ,QueueDefinitionType[0]);
         Aprintf(LOGBUF);
         *pFeedbackCode = XRMA_FB_QWRONGTYPE + QueueDefinitionType[0];
         rc = FAILED;
       }
     }
  }
  
  /*******************************************************************/
  /* Put a persistent message, with the CorrelId = ObjectInstanceId, */
  /* to the state queue.                                             */
  /*******************************************************************/
  if (rc == OK)
  {
    sprintf(LOGBUF,MSGPUTTOSTATEQMSG);
    Aprintf(LOGBUF); 
   
    pmo.Options = MQPMO_SYNCPOINT;

    memcpy(md.CorrelId
          ,pMQRMH -> ObjectInstanceId
          ,sizeof(md.CorrelId)
          ); 

    md.Persistence = MQPER_PERSISTENT;

    MQPUT(pSaveData -> HConn
         ,Hobj
         ,&md
         ,&pmo
         ,0
         ,NULL_POINTER
         ,&CompCode
         ,&Reason
         );      
              
    if (CompCode != MQCC_OK)
    {
      sprintf(LOGBUF,MQPUTFAILEDMSG, StateQueueName, Reason);
      Aprintf(LOGBUF);
      *pFeedbackCode = XRMA_FB_MQPUTSTATEFAILED + Reason;
      rc = FAILED;
    }
   
    /******************************************************************/
    /* Close the queue (if open)                                      */
    /******************************************************************/
    if (Hobj != MQHO_UNUSABLE_HOBJ)
    {
      MQCLOSE(pSaveData -> HConn
             ,&Hobj
             ,co
             ,&CompCode
             ,&Reason  
             );

      if (Reason != MQRC_NONE)
      {
        sprintf(LOGBUF,MQCLOSEFAILEDMSG,StateQueueName,Reason);
        Aprintf(LOGBUF);
      }  
    } 
  }

  return rc;
}

#if MQAT_DEFAULT == MQAT_UNIX
 void MQStart(){;}
#endif
