 /********************************************************************/
 /*                                                                  */
 /* Program name: RMSGPUT                                            */
 /*                                                                  */
 /* Description: Sample C program that creates a reference message,  */
 /*              referring to a file,and puts it to a queue.         */
 /*                                                                  */
 /*  Statement:     Licensed Materials - Property of IBM             */ 
 /*                                                                  */ 
 /*                 84H2000, 5765-B73                                */
 /*                 84H2001, 6539-B42                                */ 
 /*                 84H2002, 5765-B74                                */
 /*                 84H2003, 5765-B75                                */
 /*                 84H2004, 6539-B43                                */ 
 /*                 94H5358, 5639-B69                                */
 /*                 (C) Copyright IBM Corp. 1994, 1997               */
 /*                                                                  */
 /********************************************************************/
 /*                                                                  */
 /* Function:                                                        */
 /*                                                                  */
 /*                                                                  */
 /*   RMSGPUT is a sample C program that creates a reference         */
 /*   message that refers to a file and puts it to a specified queue.*/
 /*                                                                  */
 /*      -- The destination file can have a name different from      */
 /*         that of the source file.                                 */
 /*                                                                  */
 /*      -- The Format of the data in the file is assumed to be      */
 /*         MQFMT_STRING.                                            */
 /*                                                                  */
 /*      -- The reference message is non-persistent.                 */
 /*                                                                  */
 /*    Program logic:                                                */
 /*         Parse and validate the input parameters.                 */
 /*         MQCONN to local queue manager                            */
 /*         Create the reference message.                            */
 /*         MQPUT1 the reference message to target queue             */
 /*         MQDISC from queue manager                                */
 /*                                                                  */
 /********************************************************************/
 /*                                                                  */
 /* The following definitions will allow a reference message to be   */
 /* built to send file 'd:\file.in' with an object type of           */
 /* FLATFILE from queue manager QMGR1 to QMGR2 and to re-create      */
 /* the file as 'd:\file.out'.                                       */
 /*                                                                  */
 /* The reference message is built and put to the remote queue       */
 /* SAMPLE.MARS.REMOTE that resides in the MARS queue manager.       */
 /*                                                                  */
 /*   rmsgput -q SAMPLE.MARS.REMOTE -m MARS -i d:\x\file.in          */
 /*           -o d:\y\file.out  -t FLATFILE                          */
 /*                                                                  */
 /********************************************************************/
 /*                                                                  */
 /*   rmsgput has the following parameters                           */
 /*                                                                  */
 /*      -m queue-mgr-name     Name of local queue manager           */
 /*                            (optional).                           */
 /*                            Default is the default queue manager. */
 /*                                                                  */
 /*      -i source-file        Fully qualified name of source file   */
 /*                            to be transferred (required).         */
 /*                            The name is limited to 256 characters */
 /*                            but this can easily be changed.       */
 /*                                                                  */
 /*      -o target-file        Fully qualified name of file on the   */
 /*                            destination systems (optional).       */
 /*                            The name is limited to 256 characters */
 /*                            but this can easily be changed.       */
 /*                            Default is the source file name.      */
 /*                                                                  */
 /*      -q queue-name         Destination queue to which the        */
 /*                            reference message is put (required).  */
 /*                                                                  */
 /*      -g queue-mgr-name     Queue manager on which queue, named   */
 /*                            in -q parameter, exists (optional).   */
 /*                            Defaults to the queue manager         */
 /*                            specified by the -m parameter or the  */
 /*                            default queue manager.                */
 /*                                                                  */
 /*      -t object-type        Object type (required).               */
 /*                            Limited to 8 characters.              */
 /*                                                                  */
 /********************************************************************/
 
 #include <stdio.h>
 #include <stddef.h>
 #include <stdlib.h>
 #include <string.h> 
 #include <ctype.h>
    /* includes for MQI */
 #include <cmqc.h>

 /********************************************************************/
 /* Constants                                                        */
 /********************************************************************/

 #define PRM_OPTIONS "g:i:m:o:q:t:"
 #define NULL_POINTER (void*)0L
 #define COLON (':')
 #define BADCH ('?')
 #define TRUE  1
 #define FALSE 0
 #define MAX_FILENAME_LENGTH 256
 #define MAX_MQRMH_LENGTH    1000
 #define OK     0
 #define FAILED 1

 /********************************************************************/
 /* typedefs                                                         */
 /********************************************************************/
 typedef struct tagMQRMHX{
  MQRMH  ref;
  MQCHAR SrcName[MAX_FILENAME_LENGTH];
  MQCHAR DestName[MAX_FILENAME_LENGTH];
 } MQRMHX;

 /********************************************************************/
 /* Error and information messages                                   */
 /********************************************************************/

 #define STARTMSG              "AMQSPRM starting\n"
 #define ENDMSG                "AMQSPRM ending\n"
 #define UNRECOGNISEDPARMMSG   "Unrecognisable parameter %c\n"
 #define FILENAMETOOLONGMSG    "File name %s is too long\n"
 #define QUEUENAMETOOLONGMSG   "Queue name %s is too long\n"
 #define OBJECTTYPETOOLONGMSG  \
                 "Object type %s is too long. It has been truncated\n"
 #define QMGRNAMETOOLONGMSG    "Queue manager name %s is too long\n"
 #define DESTINATIONQUEUEMSG   "Destination queue is %s\n"
 #define DESTINATIONQMGRMSG    "Destination queue manager is %s\n"
 #define NOSOURCEFILEMSG       "No source file name specified\n"
 #define NODESTINATIONMSG      "No destination queue name specified\n"
 #define NOOBJECTTYPEMSG       "No object type specified\n"
 #define MQCONNREASONMSG       "MQCONN failed with reason %d\n"
 #define MQPUT1REASONMSG       "MQPUT1 failed with reason %d\n"
 #define MQDISCREASONMSG       "MQDISC failed with reason %d\n"
 #define SOURCEFILEMSG         "Source file is %s\n"
 #define DESTFILEMSG           "Destination file is %s\n"
 #define OBJECTTYPEMSG         "Object type is %.8s\n"
                                                                        
 /********************************************************************/
 /* Function prototypes                                              */
 /********************************************************************/
 int prmGetOpt(int    argc
              ,char **argv
              ,char  *ostr
              );

 /********************************************************************/
 /* Global variables                                                 */
 /********************************************************************/
 int prmind = 1;          /* index into argv vector                  */
 int prmopt;              /* character checked for validity          */
 char * prmarg;           /* argument associated with option         */    
 
 int main(int argc, char **argv)
 {
   /*   Declare MQI structures needed                                */
   MQOD     od = {MQOD_DEFAULT};    /* Object Descriptor             */
   MQMD     md = {MQMD_DEFAULT};    /* Message Descriptor            */
   MQPMO    pmo = {MQPMO_DEFAULT};  /* put message options           */
   MQRMHX   refx = {{MQRMH_DEFAULT}}; /* reference message           */
      /** note, sample uses defaults where it can **/

   MQHCONN  Hcon;                   /* connection handle             */
   MQLONG   CompCode = MQCC_OK;     /* completion code               */
   MQLONG   ConnCode = MQCC_FAILED; /* MQCONN completion code        */
   MQLONG   Reason;                 /* reason code                   */
   MQLONG   CReason;                /* reason code for MQCONN        */
   PMQRMH   pMQRMH = NULL_POINTER;  /* -> MQRMH structure            */
   char     QMName[MQ_Q_MGR_NAME_LENGTH+1] = "";
                                    /* local queue manager name      */
   char     DestQName[MQ_Q_NAME_LENGTH+1];
                                    /* destination queue name        */
   char     DestQMName[MQ_Q_MGR_NAME_LENGTH+1];
                                    /* destination qmgr name         */
   char     SourceFileName[MAX_FILENAME_LENGTH+1];
                                    /* source file name              */
   char     TargetFileName[MAX_FILENAME_LENGTH+1];
                                    /* target file name              */
   char     ObjectType[sizeof(refx.ref.ObjectType)];
                                    /* Object type                   */
   int      c;                      /* input option                  */
   MQLONG   ioption = FALSE;        /* /i option specified           */
   MQLONG   ooption = FALSE;        /* /o option specified           */
   MQLONG   qoption = FALSE;        /* /q option specified           */
   MQLONG   moption = FALSE;        /* /m option specified           */
   MQLONG   goption = FALSE;        /* /g option specified           */
   MQLONG   toption = FALSE;        /* /t option specified           */
   MQLONG   Arrived = FALSE;        /* Message has arrived at dest.  */ 

   /******************************************************************/
   /*                                                                */
   /* Initialisation                                                 */
   /*                                                                */
   /******************************************************************/ 
    
   printf(STARTMSG);
  
   /******************************************************************/
   /*                                                                */
   /*   Parse input parameters                                       */
   /*                                                                */
   /******************************************************************/
   while ((c = prmGetOpt(argc,argv,PRM_OPTIONS)) != EOF)
   {
     switch (c)
     {
       case 'g':
         if (strlen(prmarg) > MQ_Q_MGR_NAME_LENGTH)
         {
           printf(QMGRNAMETOOLONGMSG,prmarg);
           CompCode = MQCC_FAILED;
         }
         memset(DestQMName,0,sizeof(DestQMName)); 
         strncpy(DestQMName,prmarg
                ,(size_t)MQ_Q_MGR_NAME_LENGTH);
         printf(DESTINATIONQMGRMSG,prmarg);
         goption = TRUE;
         break;
       case 'i':
         if (strlen(prmarg) > MAX_FILENAME_LENGTH)
         {
           printf(FILENAMETOOLONGMSG,prmarg);
           CompCode = MQCC_FAILED;
         }
         memset(SourceFileName,0,sizeof(SourceFileName)); 
         strncpy(SourceFileName,prmarg,MAX_FILENAME_LENGTH);
         printf(SOURCEFILEMSG,prmarg);
         ioption = TRUE;
         break;
       case 'm':
         if (strlen(prmarg) > MQ_Q_MGR_NAME_LENGTH)
         {
           printf(QMGRNAMETOOLONGMSG,prmarg);
           CompCode = MQCC_FAILED;
         } 
         memset(QMName,0,sizeof(QMName)); 
         strncpy(QMName,prmarg
                ,(size_t)MQ_Q_MGR_NAME_LENGTH);
         moption = TRUE;
         break;
       case 'q':
         if (strlen(prmarg) > MQ_Q_NAME_LENGTH)
         {
           printf(QUEUENAMETOOLONGMSG,prmarg);
           CompCode = MQCC_FAILED;
         } 
         memset(DestQName,0,sizeof(DestQName)); 
         strncpy(DestQName,prmarg
                ,(size_t)MQ_Q_NAME_LENGTH);
         printf(DESTINATIONQUEUEMSG,prmarg);
         qoption = TRUE;
         break;
       case 'o':
         if (strlen(prmarg) > MAX_FILENAME_LENGTH)
         {
           printf(FILENAMETOOLONGMSG,prmarg);
           CompCode = MQCC_FAILED;
         } 
         memset(TargetFileName,0,sizeof(TargetFileName)); 
         strncpy(TargetFileName,prmarg,MAX_FILENAME_LENGTH);
         printf(DESTFILEMSG,prmarg);
         ooption = TRUE;
         break;
       case 't':
         if (strlen(prmarg) > sizeof(ObjectType))
         {
           printf(OBJECTTYPETOOLONGMSG,prmarg);
         }
         memset(ObjectType,' ',sizeof(ObjectType)); 
         memcpy(ObjectType
               ,prmarg
               ,(strlen(prmarg) > sizeof(ObjectType))
                ? sizeof(ObjectType)
                : strlen(prmarg)
               );
         printf(OBJECTTYPEMSG,ObjectType);
         toption = TRUE;
         break;
       default:
         printf(UNRECOGNISEDPARMMSG,prmopt);
         CompCode = MQCC_WARNING;
         break;
      } 
   }

   if (!ioption)
   {
     /****************************************************************/
     /* Source filename (option /i) must be specified                */ 
     /****************************************************************/
     printf(NOSOURCEFILEMSG);
     CompCode = MQCC_FAILED;
   }

   if (!goption)
   {
     /****************************************************************/
     /* Destination queue manager name (option /g) not specified.    */
     /* Default to local queue manager.                              */ 
     /****************************************************************/
     memcpy(DestQMName,QMName,sizeof(QMName));
     printf(DESTINATIONQMGRMSG,DestQMName);
   }

   if (!ooption)
   {
     /****************************************************************/
     /* Destination filename (option /o) not specified.              */
     /* Default to source name.                                      */ 
     /****************************************************************/
     strncpy(TargetFileName,SourceFileName,strlen(SourceFileName));
     printf(DESTFILEMSG,TargetFileName);
   }
   if (!qoption)
   {
     /****************************************************************/
     /* Destination queue name (option /q) must be specified.        */ 
     /****************************************************************/
     printf(NODESTINATIONMSG);
     CompCode = MQCC_FAILED;   
   } 
   
   if (!toption)
   {
     /****************************************************************/
     /* Object type (option /t) must be specified                    */
     /****************************************************************/
     printf(NOOBJECTTYPEMSG);
     CompCode = MQCC_FAILED;   
   }

   /******************************************************************/
   /*                                                                */
   /*   Connect to queue manager                                     */
   /*                                                                */
   /******************************************************************/
   if (CompCode != MQCC_FAILED)
   {
     MQCONN(QMName                 /* queue manager                  */
           ,&Hcon                  /* connection handle              */
           ,&ConnCode              /* completion code                */
           ,&CReason);             /* reason code                    */
                         
     CompCode = ConnCode;
                    
     if (CompCode == MQCC_FAILED)
     {
       printf(MQCONNREASONMSG, CReason);
     }
   }


   /******************************************************************/
   /*                                                                */
   /*   Build reference message.                                     */
   /*   Set Format to MQFMT_STRING.                                  */
   /*                                                                */
   /******************************************************************/
   if (CompCode != MQCC_FAILED)
   {
     refx.ref.StrucLength         = sizeof(refx);
     refx.ref.Encoding            = MQENC_NATIVE;
     memcpy(refx.ref.Format,MQFMT_STRING
           ,(size_t)MQ_FORMAT_LENGTH);
     refx.ref.Flags               = MQRMHF_LAST;
     memcpy(refx.ref.ObjectType,ObjectType,
            sizeof(refx.ref.ObjectType));

     memset(refx.SrcName
           ,' '
           ,sizeof(refx.SrcName)+sizeof(refx.DestName));

     memcpy(refx.SrcName
           ,SourceFileName
           ,strlen(SourceFileName)
           );
     memcpy(refx.DestName
           ,TargetFileName
           ,strlen(TargetFileName)
           );

     refx.ref.SrcNameLength = strlen(SourceFileName);
     refx.ref.SrcNameOffset = offsetof(MQRMHX,SrcName);

     refx.ref.DestNameLength = strlen(TargetFileName);
     refx.ref.DestNameOffset = offsetof(MQRMHX,DestName);
   }

   /******************************************************************/
   /*                                                                */
   /*  Put the reference message to the destination queue.           */
   /*                                                                */
   /******************************************************************/
   if (CompCode != MQCC_FAILED)
   {
     /****************************************************************/
     /* Set up the object descriptor.                                */
     /****************************************************************/
     strncpy(od.ObjectName
            ,DestQName
            ,sizeof(od.ObjectName)
            );
     strncpy(od.ObjectQMgrName
            ,DestQMName
            ,sizeof(od.ObjectQMgrName)
            );
 
     pmo.Options = MQPMO_FAIL_IF_QUIESCING;
     memcpy(md.Format,MQFMT_REF_MSG_HEADER
           ,(size_t)MQ_FORMAT_LENGTH);

	 md.Persistence = MQPER_NOT_PERSISTENT; //MQPER_PERSISTENT;

     MQPUT1(Hcon,                    /* connection handle            */
            &od,                     /* object descriptor for queue  */
            &md,                     /* message descriptor           */
            &pmo,                    /* options                      */
            sizeof(refx),            /* buffer length                */
            &refx,                   /* buffer                       */ 
            &CompCode,               /* MQPUT1 completion code       */
            &Reason);                /* reason code                  */

     
     if (Reason != MQRC_NONE)
     {
        printf(MQPUT1REASONMSG, Reason);
     }
   }

   /******************************************************************/
   /*                                                                */
   /*   Disconnect from MQM if not already connected                 */
   /*                                                                */
   /******************************************************************/
   if (ConnCode != MQCC_FAILED &&
       CReason != MQRC_ALREADY_CONNECTED)
   {
     MQDISC(&Hcon,                   /* connection handle            */
            &CompCode,               /* completion code              */
            &Reason);                /* reason code                  */

                          
     if (Reason != MQRC_NONE)
     {
       printf(MQDISCREASONMSG, Reason);
     }
   }

   /******************************************************************/
   /*                                                                */
   /* END OF RMSGPUT                                                 */
   /*                                                                */
   /******************************************************************/
   printf(ENDMSG);

   switch (CompCode)
   {
     case MQCC_OK :
       CompCode = 0;
       break;
     case MQCC_WARNING :
       CompCode = 10;
       break;
     case MQCC_FAILED :
       CompCode = 20;
       break;
     default :
       CompCode = 20;
   }
   return((int)CompCode);
 }
 
 /********************************************************************/
 /* prmGetOpt                                                        */
 /*------------------------------------------------------------------*/
 /* Returns option found in argv, or EOF or BADCH(invalid option).   */
 /* prmopt is the current option which is equal to the returned      */
 /*        value except when BADCH or EOF are returned.              */
 /* prmarg is the value of the current option or NULL if the option  */
 /*        has no value                                              */
 /* prmind is the current index into the argv vector                 */ 
 /********************************************************************/
 int prmGetOpt(int argc, char **argv, char *ostr)
 {
   static char *place = "";              /* option letter processing */
   char *oli = NULL_POINTER;             /* option letter list index */
   long   i  ;                           /* loop index               */
   int   found = 0;                      /* found flag               */


   if (!*place)
   {                                      /* update scanning pointer */
     place = argv[prmind];
     /****************************************************************/
     /* Check for end of input                                       */
     /****************************************************************/

     if ((prmind >= argc) ||
         ((*place != '-') && (*place != '/')) ||
         (!*++place))
     {
        return EOF;
     }
   }

   /******************************************************************/
   /* option letter okay?                                            */
   /******************************************************************/

   prmopt = (int)*place++;            /* Current option              */

   for (i=0; ((size_t)i < strlen(ostr)) && !found; i++)
   {                            /* check in option string for option */
     if ((prmopt == (int)ostr[i]) ||
         (prmopt == (int)toupper(ostr[i])))      /* case insensitive */
     {
       found = 1;                     /* option is present           */
       oli = &ostr[i];                /* set oli to point to it      */
     }
   } /* End check in option string for option */

   if (prmopt == (int)COLON || !oli)

   {                                  /* Current option not valid    */
      if (!*place)
      {
        ++prmind;
      }

      return BADCH;
   } /* End current option not valid */

   /******************************************************************/
   /* Check if this option requires an argument                      */
   /******************************************************************/

   if (*++oli != COLON)
   {                           /* this option doesn't require an arg */
     prmarg = NULL_POINTER;
     if (!*place)
     {
        ++prmind;
     }
   }
   else
   {                            /* this option needs an argument     */
     if (*place)
     {
        prmarg = place;            /* parm follows arg directly      */
        if (*prmarg == COLON)
          prmarg++;                /* Remove colon if present        */
     }
     else
     {
       if (argc <= ++prmind)
       {                          /* No argument found               */
         place = "";
         return BADCH;
       }
       else
       {
         prmarg = argv[prmind];   /* white space                     */
         if (*prmarg == COLON)
           prmarg++;              /* Remove colon if present         */
       }
     }

     place = "";
     ++prmind;
   } /* End this option needs an argument */

   prmopt = tolower(prmopt);

   return prmopt;                 /* return option letter            */
 }
